from loguru import logger
import polars as pl
import pydre.core
from pydre.metrics import registerMetric
import numpy as np
import math
from scipy import signal


# These metrics were used in driving distraction evaluation. They have not been fully tested after conversion from
# Pandas to Polars.


@registerMetric()
def getTaskNum(drivedata: pydre.core.DriveData):
    required_col = ["TaskNum"]
    drivedata.checkColumnsNumeric(required_col)
    taskNum = drivedata.data.get_column("TaskNum").mode()
    if len(taskNum) > 0:
        return taskNum[0]
    else:
        return None


@registerMetric("errorPresses")
def numOfErrorPresses(drivedata: pydre.core.DriveData):
    required_col = ["SimTime", "TaskFail"]
    drivedata.checkColumnsNumeric(required_col)
    df = drivedata.data.select(required_col)
    df = df.unique(maintain_order=True).drop_nulls()
    presses = df.select((pl.col("TaskFail").shift() != pl.col("TaskFail")) > 0).sum()
    return presses


def appendDFToCSV_void(df, csvFilePath: str, sep: str = ","):
    if not os.path.isfile(csvFilePath):
        df.to_csv(csvFilePath, mode="a", index=False, sep=sep)
    elif len(df.columns) != len(pl.read_csv(csvFilePath, nrows=1, sep=sep).columns):
        raise Exception(
            "Columns do not match!! Dataframe has "
            + str(len(df.columns))
            + " columns. CSV file has "
            + str(len(pl.read_csv(csvFilePath, nrows=1, sep=sep).columns))
            + " columns."
        )
    elif not (df.columns == pl.read_csv(csvFilePath, nrows=1, sep=sep).columns).all():
        raise Exception(
            "Columns and column order of dataframe and csv file do not match!!"
        )
    else:
        df.to_csv(csvFilePath, mode="a", index=False, sep=sep, header=False)


@registerMetric(
    columnnames=[
        "numOfGlancesOR",
        "numOfGlancesOR2s",
        "meanGlanceORDuration",
        "sumGlanceORDuration",
    ]
)
def gazeNHTSA(drivedata: pydre.core.DriveData):
    required_col = ["VidTime", "gaze", "gazenum", "TaskFail", "taskblocks", "PartID"]
    drivedata.checkColumns(required_col)

    numofglances = 0

    df = drivedata.data.select(required_col).unique(maintain_order=True).drop_nulls()

    # construct table with columns [glanceduration, glancelocation, error]
    gr = df.group_by("gazenum")

    glancelist = gr.agg(
        durations=(pl.col("VidTime").max() - pl.col("VidTime").min()),
        locations=pl.col("gaze")
        .first()
        .fill_null("offroad")
        .replace(
            ["car.WindScreen", "car.dashPlane", "None"],
            ["onroad", "offroad", "offroad"],
        ),
        error_list=pl.col("TaskFail").any(),
        TaskID=pl.col("TaskID").min(),
        taskblock=pl.col("taskblocks").min(),
        Subject=pl.col("PartID").min(),
    )

    # appendDFToCSV_void(glancelist_aug, "glance_list.csv")

    # table constructed, now find metrics
    # glancelist['over2s'] = glancelist['duration'] > 2

    num_over_2s_offroad_glances = glancelist[
        (glancelist["duration"] > 2) & (glancelist["locations"] == "offroad")
    ]["duration"].count()

    num_offroad_glances = glancelist[(glancelist["locations"] == "offroad")][
        "duration"
    ].count()

    total_time_offroad_glances = glancelist[(glancelist["locations"] == "offroad")][
        "duration"
    ].sum()

    mean_time_offroad_glances = glancelist[(glancelist["locations"] == "offroad")][
        "duration"
    ].mean()

    # print(">2s glances: {}, num glances: {}, total time glances: {}, mean time glances {}".format(
    # num_over_2s_offroad_glances, num_offroad_glances, total_time_offroad_glances, mean_time_offroad_glances))

    return [
        num_offroad_glances,
        num_over_2s_offroad_glances,
        mean_time_offroad_glances,
        total_time_offroad_glances,
    ]


# not working
def addVelocities(drivedata: pydre.core.DriveData):
    df = pl.DataFrame(drivedata.data)
    # add column with ownship velocity
    g = np.gradient(df.XPos.values, df.SimTime.values)
    df.insert(len(df.columns), "OwnshipVelocity", g, True)
    # add column with leadcar velocity
    headwayDist = df.HeadwayTime * df.OwnshipVelocity
    # v = df.OwnshipVelocity+np.gradient(headwayDist, df.SimTime.values)
    df.insert(len(df.columns), "LeadCarPos", headwayDist + df.XPos.values, True)
    df.insert(len(df.columns), "HeadwayDist", headwayDist, True)
    v = np.gradient(headwayDist + df.XPos.values, df.SimTime.values)
    df.insert(len(df.columns), "LeadCarVelocity", v, True)
    return df


@registerMetric()
def crossCorrelate(drivedata: pydre.core.DriveData):
    df = drivedata.data
    if "OwnshipVelocity" not in df.columns or "LeadCarVelocity" not in df.columns:
        df = addVelocities(drivedata)
        print("calling addVelocities()")

    v2 = df.LeadCarVelocity
    v1 = df.OwnshipVelocity
    cor = signal.correlate(v1 - v1.mean(), v2 - v2.mean(), mode="same")
    # cor-An N-dimensional array containing a subset of the discrete linear cross-correlation of in1 with in2.
    # delay index at the max
    df.insert(len(df.columns), "CrossCorrelations", cor, True)
    delayIndex = np.argmax(cor)
    if delayIndex > 0:
        v2 = df.LeadCarVelocity.iloc[delayIndex : df.columns.__len__()]
        v1 = df.OwnshipVelocity.iloc[delayIndex : df.columns.__len__()]
    # normalize vectors
    v1_norm = v1 / np.linalg.norm(v1)
    v2_norm = v2 / np.linalg.norm(v2)
    cor = np.dot(v1_norm, v2_norm)
    if cor > 0:
        return cor
    else:
        return 0.0


# find relative time where speed is within [mpsBound] of new speed limit
# 0 is when the car is crossing the sign.
# Returns None if the speed is never within 2
@registerMetric()
def speedLimitMatchTime(
    drivedata: pydre.core.DriveData, mpsBound: float, speedLimitCol: str
):
    required_col = ["DatTime", speedLimitCol, "Velocity"]
    diff = drivedata.checkColumns(required_col)

    speed_limit = drivedata.data.get_column(speedLimitCol).tail(1).item() * 0.44704
    starting_speed_limit = (
        drivedata.data.get_column(speedLimitCol).head(1).item() * 0.44704
    )

    if speed_limit == 0 or starting_speed_limit == 0:
        return None

    if speed_limit > starting_speed_limit:
        # increasing speed
        match_speed_block = drivedata.data.filter(
            pl.col("Velocity") >= (speed_limit - mpsBound)
        )
    else:
        match_speed_block = drivedata.data.filter(
            pl.col("Velocity") <= (speed_limit + mpsBound)
        )

    if match_speed_block.height > 0:
        time = match_speed_block.item(0, "DatTime")
    else:
        time = drivedata.data.tail(1).get_column("DatTime").item()

    sign_time = drivedata.data.filter(
        abs(pl.col(speedLimitCol) * 0.44704 - starting_speed_limit) > 0.1
    ).item(0, "DatTime")

    if time is None:
        return None
    else:
        # print( starting_speed_limit, speed_limit, time, sign_time)
        return time - sign_time


# The following functions need to be revised to work with polars rather than pandas

#
# @registerMetric(
#     columnnames=[
#         "total_time_onroad_glance",
#         "percent_onroad",
#         "avg_offroad",
#         "avg_onroad",
#     ]
# )
# def speedbumpHondaGaze(drivedata: pydre.core.DriveData):
#     required_col = ["DatTime", "gaze", "gazenum", "TaskNum", "taskblocks", "PartID"]
#     diff = drivedata.checkColumns(required_col)
#
#     numofglances = 0
#     df = pl.DataFrame(drivedata.data, columns=required_col)  # drop other columns
#     df = pl.DataFrame.drop_duplicates(
#         df.dropna(axis=0, how="any")
#     )  # remove nans and drop duplicates
#
#     if len(df) == 0:
#         return
#
#     # construct table with columns [glanceduration, glancelocation, error]
#     gr = df.groupby("gazenum", sort=False)
#     durations = gr["DatTime"].max() - gr["DatTime"].min()
#     locations = gr["gaze"].first()
#     error_list = gr["TaskNum"].any()
#
#     glancelist = pl.DataFrame(
#         {"duration": durations, "locations": locations, "errors": error_list}
#     )
#     glancelist["locations"].fillna("offroad", inplace=True)
#     glancelist["locations"].replace(
#         ["car.WindScreen", "car.dashPlane", "None"],
#         ["onroad", "offroad", "offroad"],
#         inplace=True,
#     )
#
#     glancelist_aug = glancelist
#     glancelist_aug["TaskNum"] = drivedata.data["TaskNum"].min()
#     glancelist_aug["taskblock"] = drivedata.data["taskblocks"].min()
#     glancelist_aug["Subject"] = drivedata.data["PartID"].min()
#
#     appendDFToCSV_void(glancelist_aug, "glance_list.csv")
#
#     # table constructed, now find metrics
#
#     num_onroad_glances = glancelist[(glancelist["locations"] == "onroad")][
#         "duration"
#     ].count()
#
#     total_time_onroad_glances = glancelist[(glancelist["locations"] == "onroad")][
#         "duration"
#     ].sum()
#     percent_onroad = total_time_onroad_glances / (
#         df["DatTime"].max() - df["DatTime"].min()
#     )
#
#     mean_time_offroad_glances = glancelist[(glancelist["locations"] == "offroad")][
#         "duration"
#     ].mean()
#     mean_time_onroad_glances = glancelist[(glancelist["locations"] == "onroad")][
#         "duration"
#     ].mean()
#
#     return [
#         total_time_onroad_glances,
#         percent_onroad,
#         mean_time_offroad_glances,
#         mean_time_onroad_glances,
#     ]
#
#
# @registerMetric(
#     columnnames=["85th_percentile", "duration_mean", "duration_median", "duration_std"]
# )
# def speedbumpHondaGaze2(
#     drivedata: pydre.core.DriveData, timecolumn="DatTime", maxtasknum=5
# ):
#     required_col = [
#         timecolumn,
#         "gaze",
#         "gazenum",
#         "TaskNum",
#         "TaskFail",
#         "TaskInstance",
#         "KEY_EVENT_T",
#         "KEY_EVENT_P",
#     ]
#     # filters.numberTaskInstances() is required.
#     # for now I just assume the input dataframe has a TaskInstance column
#     # diff = drivedata.checkColumns(required_col)
#
#     logger.warning("Processing Task {}".format(drivedata.data.TaskNum.mean()))
#     if drivedata.data.TaskNum.mean() == 0 or drivedata.data.TaskNum.mean() > maxtasknum:
#         return [None, None, None, None]
#     # if d.TaskNum.mean() == 4:
#     #    d.to_csv('4.csv')
#     df = pl.DataFrame(drivedata.data, columns=required_col)  # drop other columns
#     df = df.fillna(0)  # remove nans and drop duplicates
#
#     df["time_diff"] = df[
#         timecolumn
#     ].diff()  # get durations by calling time_column.diff()
#
#     df = df.loc[df.gaze != "onroad"]  # remove onroad rows
#     # df.to_csv('AAM_cp1.csv')
#     df = df.loc[
#         (df["TaskInstance"] != 0) & (df["TaskInstance"] != np.nan)
#     ]  # drop all rows that are not in any task instance
#     dropped_instances = df.loc[(df["TaskFail"] == 1)]
#     dropped_instances = (
#         dropped_instances["TaskInstance"].drop_duplicates()
#     )  # get all the task instances that contains a fail and needs to be dropped
#     df = df.loc[
#         ~df["TaskInstance"].isin(dropped_instances)
#     ]  # drop all the failed task instances
#     # df.to_csv('AAM_cp2.csv')
#
#     # get first 8 task instances
#     number_valid_instance = df["TaskInstance"].unique()
#     print(pl.unique(df.TaskInstance))
#     if len(number_valid_instance) > 8:
#         lowest_instance_no = number_valid_instance.min()
#         len_of_drop = len(
#             dropped_instances.loc[dropped_instances < (lowest_instance_no + 8)]
#         )
#         highest_instance_no = lowest_instance_no + 8 + len_of_drop
#         df = df.loc[
#             (df["TaskInstance"] < highest_instance_no)
#             & (df["TaskInstance"] >= lowest_instance_no)
#         ]
#         # logger.warning(highest_instance_no)
#         # logger.warning(lowest_instance_no)
#     elif len(number_valid_instance) < 8:
#         logger.warning(
#             "Not enough valid task instances. Found {}".format(
#                 len(number_valid_instance)
#             )
#         )
#
#     # df = df[df.TaskInstance < 9] # only keep the glance data for the first eight task instances for each task per person.
#
#     # print(df)
#     # df.to_csv('df.csv')
#
#     group_by_gazenum = df.groupby("gazenum", sort=False)
#     durations_by_gazenum = group_by_gazenum.sum()
#     durations_by_gazenum = durations_by_gazenum.loc[
#         (durations_by_gazenum["time_diff"] != 0.0)
#     ]
#     # print(durations_by_gazenum)
#     percentile = np.percentile(
#         durations_by_gazenum.time_diff, 85
#     )  # find the 85th percentile value (A1)
#
#     group_by_instance = df.groupby("TaskInstance", sort=False)  # A2
#     durations_by_instance = group_by_instance.sum()
#     print(durations_by_instance)
#     sum_mean = (durations_by_instance.time_diff).mean()  # mean of duration sum
#     sum_median = (durations_by_instance.time_diff).median()  # median of duration sum
#     sum_std = (durations_by_instance.time_diff).std()  # std of duration sum
#
#     return [percentile, sum_mean, sum_median, sum_std]


@registerMetric()
def eventCount(drivedata: pydre.core.DriveData, event="KEY_EVENT_S"):
    required_col = [event]
    diff = drivedata.checkColumns(required_col)

    df = pl.DataFrame(drivedata.data, columns=required_col)
    col_name = event + "_ocurrance"
    df[col_name] = df[event].diff()
    occur = df[col_name].value_counts().get(1)
    if occur == None:
        occur = 0
    return occur


@registerMetric()
def insDuration(drivedata: pydre.core.DriveData):
    required_col = ["DatTime", "TaskInstance"]
    diff = drivedata.checkColumns(required_col)

    df = pl.DataFrame(drivedata.data, columns=required_col)
    return df.tail(1).iat[0, 0] - df.head(1).iat[0, 0]


@registerMetric(
    columnnames=[
        "mean_onroad_spbpon",
        "mean_offroad_spbpon",
        "mean_onroad_spbpoff",
        "mean_offroad_spbpoff",
    ]
)
def speedbump2Gaze(drivedata: pydre.core.DriveData, timecolumn="DatTime", duration=6.0):
    required_col = [
        timecolumn,
        "gaze",
        "gazenum",
        "TaskNum",
        "TaskFail",
        "KEY_EVENT_S",
        "KEY_EVENT_T",
        "KEY_EVENT_P",
    ]
    diff = drivedata.checkColumns(required_col)

    if (
        drivedata.data.TaskNum.mean() == 0
        or drivedata.data.TaskNum.mean() < 4
        or drivedata.data.TaskNum.mean() > 5
    ):
        return [None, None, None, None]

    df = pl.DataFrame(drivedata.data, columns=required_col)  # drop other columns
    # df.to_csv("df.csv")
    # df = df.loc[(df['TaskInstance'] != 0) & (df['TaskInstance'] != np.nan)] # drop all rows that are not in any task instance
    # dropped_instances = df.loc[(df['TaskFail'] == 1)]
    # dropped_instances = dropped_instances['TaskInstance'].drop_duplicates() # get all the task instances that contains a fail and needs to be dropped
    # df = df.loc[~df['TaskInstance'].isin(dropped_instances)] # drop all the failed task instances

    event_s = pl.DataFrame(df, columns=[timecolumn, "gazenum", "gaze", "KEY_EVENT_S"])
    event_s["diff"] = event_s["KEY_EVENT_S"].diff()
    event_s["s_begin"] = 0
    event_s.loc[event_s["diff"] == 1, "s_begin"] = (
        1  # mark the beginning of all the s events
    )

    s_begin_time = event_s.loc[event_s["s_begin"] == 1]
    s_begin_time = s_begin_time.reset_index()
    # print(s_begin_time)
    index = 0
    for time in s_begin_time[timecolumn]:  # mark the duration of speedbump as 1
        event_s.loc[
            (event_s[timecolumn] > float(time))
            & (event_s[timecolumn] < float(time + duration)),
            "s_begin",
        ] = 1
        index += 1
    # event_s.to_csv("e.csv")

    group_by_gazenum = event_s.groupby("gazenum", sort=False)
    durations = pl.DataFrame(group_by_gazenum.sum(), columns=["s_begin"])

    durations_begin = pl.DataFrame(group_by_gazenum.min(), columns=[timecolumn])
    durations_begin = durations_begin.rename(columns={timecolumn: "begin"})
    durations_end = pl.DataFrame(group_by_gazenum.max(), columns=[timecolumn])
    durations_end = durations_end.rename(columns={timecolumn: "end"})

    durations["begin"] = durations_begin["begin"]
    durations["end"] = durations_end[
        "end"
    ]  # contains the beginning time and end time of all gazes
    durations = durations.reset_index()
    durations["gaze"] = group_by_gazenum.min()["gaze"].tolist()

    durations["in_speedbump"] = 0

    i = 0
    while i < len(durations):
        current_gaze = durations.at[i, "gazenum"]
        begin_time = durations.at[i, "begin"]
        end_time = durations.at[i, "end"]

        if durations.at[i, "s_begin"] > 0:
            temp = event_s.loc[event_s["gazenum"] == current_gaze]
            total_length = len(temp.index)  # length of the entire glance
            speedbump_act = temp.loc[
                temp["s_begin"] == 1
            ]  # length of the time period when speedbump is on during the current glance
            act_length = len(speedbump_act.index)
            if act_length >= (
                total_length / 2
            ):  # if speedbump is on for more than half of the total glance time:
                durations.at[i, "in_speedbump"] = 1
        i += 1

    durations["gaze_duration"] = durations["end"] - durations["begin"]

    durations_onroad_spbpon = durations.loc[
        (durations["gaze"] == "onroad") & (durations["in_speedbump"] == 1)
    ]
    durations_offroad_spbpon = durations.loc[
        (durations["gaze"] == "offroad") & (durations["in_speedbump"] == 1)
    ]
    durations_onroad_spbpoff = durations.loc[
        (durations["gaze"] == "onroad") & (durations["in_speedbump"] == 0)
    ]
    durations_offroad_spbpoff = durations.loc[
        (durations["gaze"] == "offroad") & (durations["in_speedbump"] == 0)
    ]

    # durations.to_csv("dursum.csv")
    # durations_onroad_spbpon.to_csv("duron_so.csv")
    # durations_offroad_spbpon.to_csv("duroff_so.csv")
    # durations_onroad_spbpoff.to_csv("duron_sx.csv")
    # durations_offroad_spbpoff.to_csv("duroff_sx.csv")

    mean_onroad_spbpon = durations_onroad_spbpon["gaze_duration"].mean()
    mean_offroad_spbpon = durations_offroad_spbpon["gaze_duration"].mean()
    mean_onroad_spbpoff = durations_onroad_spbpoff["gaze_duration"].mean()
    mean_offroad_spbpoff = durations_offroad_spbpoff["gaze_duration"].mean()
    return [
        mean_onroad_spbpon,
        mean_offroad_spbpon,
        mean_onroad_spbpoff,
        mean_offroad_spbpoff,
    ]



from loguru import logger
import json
import pathlib
import tomlkit


def convertProjToToml(proj: dict) -> tomlkit.TOMLDocument:
    new_proj = tomlkit.TOMLDocument()
    filters = proj.get("filters", [])
    rois = proj.get("rois", [])
    metrics = proj.get("metrics", [])

    if len(filters) > 0:
        filters_table = tomlkit.table(True)
        for filter in filters:
            name = filter.pop("name", None)
            f_tab = tomlkit.table()
            f_tab.update(filter)
            filters_table.add(name, f_tab)
        new_proj.add("filters", filters_table)

    if len(rois) > 0:
        rois_table = tomlkit.table(True)
        i = 1
        for roi in rois:
            name = f"roi{i}"
            i += 1
            r_tab = tomlkit.table()
            r_tab.update(roi)
            rois_table.add(name, r_tab)
        new_proj.add("rois", rois_table)

    if len(metrics) > 0:
        metrics_table = tomlkit.table(True)
        for metric in metrics:
            name = metric.pop("name", None)
            m_tab = tomlkit.table()
            m_tab.update(metric)
            metrics_table.add(name, m_tab)
        new_proj.add("metrics", metrics_table)
    return new_proj


def main():
    proj_dir = pathlib.Path("./project_files")
    old_proj_files = proj_dir.resolve().glob("*.json")
    for fn in old_proj_files:
        logger.info(f"Converting {fn}")
        with open(fn, "r") as fs:
            try:
                json_data = json.load(fs)
            except json.decoder.JSONDecodeError:
                logger.warning(f"Error decoding {fn}")
                continue
        new_proj = convertProjToToml(json_data)
        with open(fn.with_suffix(".toml"), "w") as fs:
            tomlkit.dump(new_proj, fs)


if __name__ == "__main__":
    main()



import pytest
from loguru import logger
import logging
from _pytest.logging import LogCaptureFixture


@pytest.fixture
def caplog(caplog: LogCaptureFixture):
    handler_id = logger.add(
        caplog.handler,
        format="{message}",
        level=0,
        filter=lambda record: record["level"].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    logger.remove(handler_id)


@pytest.fixture
def reportlog(pytestconfig):
    logging_plugin = pytestconfig.pluginmanager.getplugin("logging-plugin")
    handler_id = logger.add(logging_plugin.report_handler, format="{message}")
    yield
    logger.remove(handler_id)


@pytest.fixture(autouse=True)
def propagate_logs():
    class PropagateHandler(logging.Handler):
        def emit(self, record):
            if logging.getLogger(record.name).isEnabledFor(record.levelno):
                logging.getLogger(record.name).handle(record)

    logger.remove()
    logger.add(PropagateHandler(), format="{message}")
    yield



from loguru import logger
from . import project
import sys
import pathlib
import argparse


def main():
    parser = argparse.ArgumentParser()
    # command line arguments for project file (pf) and data file (df)
    parser.add_argument(
        "-p", "--projectfile", type=str, help="the project file path", required=True
    )
    parser.add_argument(
        "-d",
        "--datafiles",
        type=str,
        help="the data file path",
        nargs="+",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--outputfile",
        type=str,
        default="out.csv",
        help="the name of the output file",
    )
    parser.add_argument(
        "-l",
        "--warninglevel",
        type=str,
        default="WARNING",
        help="Loggging error level. DEBUG, INFO, WARNING, ERROR, and CRITICAL are allowed.",
    )
    args = parser.parse_args()
    logger.remove(0)
    try:
        logger.add(sys.stderr, level=args.warninglevel.upper())
    except Exception:
        logger.add(sys.stderr, level="WARNING")
        logger.warning("Command line log level (-l) invalid. Defaulting to WARNING")
    if args.outputfile == "out.csv":
        logger.warning("No output file specified. Defaulting to 'out.csv'")
    p = project.Project(args.projectfile)
    # test the data files
    filelist = []
    for fn in args.datafiles:
        # convert relative path to absolute path
        datapath = pathlib.Path(fn).resolve()
        datafiles = datapath.parent.glob(datapath.name)
        filelist.extend(datafiles)
    p.processDatafiles(filelist, 12)
    p.saveResults(pathlib.Path(args.outputfile))


if __name__ == "__main__":
    main()



from loguru import logger
from typing import Optional, Callable

__all__ = ["common", "eyetracking", "R2D"]

import pydre.core

filtersList = {}
filtersColNames = {}


def registerFilter(filtername: Optional[str] = None) -> Callable:
    def registering_decorator(
        func: Callable[[pydre.core.DriveData, ...], pydre.core.DriveData],
    ) -> Callable[[pydre.core.DriveData, ...], pydre.core.DriveData]:
        name = filtername
        if not name:
            name = func.__name__
        # register function
        filtersList[name] = func
        return func

    return registering_decorator



import datetime
import struct
from loguru import logger
from pathlib import Path
import polars as pl
import pydre.core
from typing import Optional
from . import registerFilter


@registerFilter()
def numberBinaryBlocks(
    drivedata: pydre.core.DriveData,
    binary_column="ButtonStatus",
    new_column="NumberedBlocks",
    only_on=0,
    limit_fill_null=700,
    extend_blocks=0,
) -> pydre.core.DriveData:
    """Adds a column that separates data into blocks based on the value of another column

    If only_on is set to 1, it filters the data to only include rows where binary_col is set to 1.
    If extend_blocks is set to 1, it extends the blocks.

    Parameters:
        binary_column: The name of the column to reference
        new_column: The name of the new column with blocks
        only_on: Determines whether to filter the data after adding blocks.
        extend_blocks: Determines whether to extend the blocks.
        limit_fill_null: Determines how many rows to fill using fill_null (only applies when extend_blocks is set to 1).

    Returns:
        Original drive data object augmented with new column
    """

    required_col = [binary_column]
    drivedata.checkColumns(required_col)

    blocks = drivedata.data.select(
        (pl.col(binary_column).shift() != pl.col(binary_column))
        .cum_sum()
        .alias(new_column)
    )

    drivedata.data.hstack(blocks, in_place=True)
    if only_on:
        try:
            drivedata.data = drivedata.data.filter(pl.col(binary_column) == 1)
        except pl.exceptions.ComputeError as e:
            logger.warning(
                "Assumed binary column {} in {} has non-numeric value.".format(
                    binary_column, drivedata.sourcefilename
                )
            )

    if extend_blocks:
        drivedata.data = drivedata.data.with_columns(
            pl.when(pl.col(binary_column) == 0)
            .then(None)
            .otherwise(pl.col(new_column))
            .alias(new_column)
        )
        drivedata.data = drivedata.data.with_columns(
            pl.col(new_column).fill_null(strategy="forward", limit=limit_fill_null)
        )
        drivedata.data = drivedata.data.filter(pl.col(new_column).is_not_null())

    return drivedata


@registerFilter()
def SimTimeFromDatTime(drivedata: pydre.core.DriveData) -> pydre.core.DriveData:
    """Copies DatTime to SimTime

    Note: Requires data columns
        - SimTime: simulation time
        = DatTime: time from simobserver recording start

    Returns:
        Original DriveData object with identical DatTime and SimTime
    """
    drivedata.data = drivedata.data.with_columns(pl.col("DatTime").alias("SimTime"))
    return drivedata


@registerFilter()
def FixReversedRoadLinearLand(drivedata: pydre.core.DriveData) -> pydre.core.DriveData:
    """Fixes a section of reversed road in the LinearLand map

    RoadOffset becomes -RoadOffset between XPos 700 and 900

    Note: Requires data columns
        - XPos: X position of ownship
        - RoadOffset: lateral distance on roadway

    Returns:
        Original DriveData object with altered RoadOffset column data
    """
    drivedata.data = drivedata.data.with_columns(
        pl.when(pl.col("XPos").cast(pl.Float32).is_between(700, 900))
        .then(-(pl.col("RoadOffset").cast(pl.Float32)))
        .otherwise(pl.col("RoadOffset").cast(pl.Float32))
        .alias("RoadOffset")
    )
    return drivedata


@registerFilter()
def setinrange(
    drivedata: pydre.core.DriveData,
    coltoset: str,
    valtoset: float,
    colforrange: str,
    rangemin: float,
    rangemax: float,
) -> pydre.core.DriveData:
    """Set values of one column based on the values of another column

    If the value of *colforrange* is outside the range of (*rangemin*, *rangemax*), then
    the value of *coltoset* will be unchanged. Otherwise, the value of *coltoset* will be changed to *valtoset*.

    Parameters:
        coltoset: The name of the column to modify
        valtoset: The new value to set for the
        colforrange: The name of the column to look up to decide to set a new value or not
        rangemin: Minimum value of the range
        rangemax: Maximum value of the range

    Returns:
        Original DriveData object with modified column
    """
    drivedata.data = drivedata.data.with_columns(
        pl.when(pl.col(colforrange).cast(pl.Float32).is_between(rangemin, rangemax))
        .then(valtoset)
        .otherwise(pl.col(coltoset))
        .cast(pl.Float32)
        .alias(coltoset)
    )

    return drivedata


@registerFilter()
def relativeBoxPos(drivedata: pydre.core.DriveData) -> pydre.core.DriveData:
    start_x = drivedata.data.get_column("XPos").min()
    drivedata.data = drivedata.data.with_columns(
        [
            (pl.col("BoxPosY").cast(pl.Float32) - start_x)
            .clip(lower_bound=0)
            .alias("relativeBoxStart")
        ]
    )
    return drivedata


@registerFilter()
def zscoreCol(
    drivedata: pydre.core.DriveData, col: str, newcol: str
) -> pydre.core.DriveData:
    """Transform a column into a standardized z-score column

    Parameters:
        col: The name of the column to transform
        newcol: The name of the new z-score column

    Returns:
        Original DriveData object augmented with new z-score column
    """
    colMean = drivedata.data.get_column(col).mean()
    colSD = drivedata.data.get_column(col).std()
    drivedata.data = drivedata.data.with_columns(
        ((pl.col(col) - colMean) / colSD).alias(newcol)
    )
    return drivedata


@registerFilter()
def speedLimitTransitionMarker(
    drivedata: pydre.core.DriveData, speedlimitcol: str
) -> pydre.core.DriveData:
    speedlimitpos = drivedata.data.select(
        [
            (pl.col(speedlimitcol).shift() != pl.col(speedlimitcol)).alias(
                "SpeedLimitPositions"
            ),
            speedlimitcol,
            "XPos",
            "DatTime",
        ]
    )

    block_marks = speedlimitpos.filter(pl.col("SpeedLimitPositions"))
    drivedata.data = drivedata.data.with_columns(
        pl.lit(None).cast(pl.Int32).alias("SpeedLimitBlocks")
    )

    mph2mps = 0.44704

    blocknumber = 1
    for row in block_marks.rows(named=True):
        drivedata.data = drivedata.data.with_columns(
            pl.when(
                pl.col("DatTime")
                .cast(pl.Float32)
                .is_between(row["DatTime"] - 5, row["DatTime"] + 5)
            )
            .then(blocknumber)
            .otherwise(pl.col("SpeedLimitBlocks"))
            .alias("SpeedLimitBlocks")
        )
        blocknumber += 1

    return drivedata


@registerFilter()
def writeToCSV(
    drivedata: pydre.core.DriveData, outputDirectory: str
) -> pydre.core.DriveData:
    sourcefilename = Path(drivedata.sourcefilename).stem
    outputfilename = Path(outputDirectory).with_stem(sourcefilename).with_suffix(".csv")
    drivedata.data.write_csv(outputfilename)
    return drivedata


def filetimeToDatetime(ft: int) -> Optional[datetime.datetime]:
    EPOCH_AS_FILETIME = 116444736000000000  # January 1, 1970 as filetime
    HUNDREDS_OF_NS = 10000000
    s, ns100 = divmod(ft - EPOCH_AS_FILETIME, HUNDREDS_OF_NS)
    try:
        result = datetime.datetime.fromtimestamp(s, tz=datetime.timezone.utc).replace(
            microsecond=(ns100 // 10)
        )
    except OSError:
        # happens when the input to fromtimestamp is outside the legal range
        result = None
    return result


def mergeSplitFiletime(hi: int, lo: int):
    return struct.unpack("Q", struct.pack("LL", lo, hi))[0]



from __future__ import annotations

import polars
from loguru import logger
from typing import List, Optional
from pathlib import Path


class DriveData:
    data: polars.dataframe
    sourcefilename: Path
    roi: Optional[str]
    format_identifier: int
    PartID: Optional[int]
    DriveID: Optional[int]
    UniqueID: Optional[int]
    scenarioName: Optional[int]
    mode: Optional[int]

    def __init__(self, data: polars.DataFrame, sourcefilename: Path):
        self.data = data
        self.sourcefilename = sourcefilename
        self.roi = None
        self.format_identifier = -1

    @classmethod
    def initV2(
        cls,
        data: polars.DataFrame,
        sourcefilename: Path,
        PartID: Optional[int],
        DriveID: Optional[int],
    ):
        obj = cls(data, sourcefilename)
        obj.PartID = PartID
        obj.DriveID = DriveID
        obj.format_identifier = 2
        return obj

    @classmethod
    def initV4(
        cls,
        data: polars.DataFrame,
        sourcefilename: Path,
        PartID: str,
        UniqueID: Optional[int],
        scenarioName: Optional[str],
        mode: Optional[str],
    ):
        obj = cls(data, sourcefilename)
        obj.PartID = PartID
        obj.UniqueID = UniqueID
        obj.scenarioName = scenarioName
        obj.mode = mode
        obj.format_identifier = 4
        return obj

    def copyMetaData(self, other: DriveData):
        self.sourcefilename = other.sourcefilename
        self.PartID = other.PartID
        if other.format_identifier == 2:
            self.DriveID = other.DriveID
        elif other.format_identifier == 4:
            self.UniqueID = other.UniqueID
            self.scenarioName = other.scenarioName
            self.mode = other.mode
        self.format_identifier = other.format_identifier

    def checkColumns(self, required_columns: List[str]) -> None:
        difference = set(required_columns) - set(list(self.data.columns))
        if len(difference) > 0:
            raise ColumnsMatchError(
                f"Columns {difference} not found.", list(difference)
            )

    def checkColumnsNumeric(self, columns: List[str]) -> None:
        """Scans data for required column. If columns do not exist or are not numeric, raise an exception.

        args:
            columns: List of required numeric columns.
        """
        non_numeric = []
        for column in columns:
            if column in self.data:
                to_check = self.data.get_column(column)
                if not to_check.dtype.is_numeric():
                    logger.info(
                        "col("
                        + column
                        + ") is not numeric in "
                        + str(self.sourcefilename)
                    )
                    non_numeric.append(column)
            else:
                logger.info(
                    column + " in " + str(self.sourcefilename) + " does not exist."
                )
                non_numeric.append(column)
        if len(non_numeric) > 0:
            raise ColumnsMatchError(f"Columns {non_numeric} not numeric.", non_numeric)


## TODO: Update this to polars style.
# def mergeBySpace(tomerge: List[DriveData]) -> DriveData:
#     """
#     args:
#         tomerge: list of DriveData objects to merge
#
#     returns:
#         DriveData object containing merged data and Drive ID of first element in the tomerge list
#
#     Takes the first DriveData in 'tomerge', finds the last X and Y position, matches the closest X and Y
#     position in the next DriveData object in 'tomerge'.  The time stamps for the second data list are
#     altered appropriately.
#     This repeats for all elements in 'tomerge'.
#     """
#     out_frame = tomerge[0].data
#
#     if len(tomerge) > 1:
#         i = 0
#         while i < len(tomerge) - 1:
#             i = i + 1
#
#             # This part does the actual merging
#             last_line = out_frame.tail(1)
#             last_x = last_line.XPos.iloc[0]
#             last_y = last_line.YPos.iloc[0]
#             last_time = last_line.SimTime.iloc[0]
#             next_frame = tomerge[i].data
#             min_dist = float("inf")
#             min_index = 0
#             for index, row in next_frame.iterrows():
#                 dist = (row.XPos - last_x) ** 2 + (row.YPos - last_y) ** 2
#                 if dist < min_dist:
#                     min_index = index
#                     min_dist = dist
#             start_time = next_frame.iloc[min_index].SimTime
#             next_frame = next_frame[min_index:]
#             next_frame.SimTime += last_time - start_time
#             out_frame = out_frame.append(next_frame)
#
#     new_dd = DriveData(out_frame, "")
#     new_dd.copyMetaData(tomerge[0])
#     return new_dd
#


# ------ exception handling ------


class ColumnsMatchError(Exception):
    """Exception when a filter or metric expects a certain column in DriveData but it is not present or an unexpected type"""

    default_message = "Columns in DriveData object not as expected."

    def __init__(self, message: str, missing_columns: list[str]):
        super().__init__(message or f"{self.default_message} {missing_columns}")
        self.missing_columns = missing_columns



import json

import polars as pl
import re
import sys
import tomllib
from typing import Optional
import pydre.core
import pydre.rois
import pydre.metrics
from pydre.metrics import *
import pydre.filters
from pydre.filters import *
import pathlib
from pathlib import Path
from loguru import logger
from tqdm import tqdm
import concurrent.futures


class Project:
    project_filename: Path  # used only for information
    definition: dict
    results: Optional[pl.DataFrame]

    def __init__(self, projectfilename: str):
        self.project_filename = pathlib.Path(projectfilename)
        self.definition = {}
        self.results = None
        try:
            with open(self.project_filename, "rb") as project_file:
                if self.project_filename.suffix == ".json":
                    try:
                        self.definition = json.load(project_file)
                    except json.decoder.JSONDecodeError as e:
                        logger.exception(
                            "Error parsing JSON in {}".format(self.project_filename),
                            exception=e,
                        )
                        # exited as a general error because it is seemingly best suited for the problem encountered
                        sys.exit(1)
                elif self.project_filename.suffix == ".toml":
                    try:
                        self.definition = tomllib.load(project_file)
                    except tomllib.TOMLDecodeError as e:
                        logger.exception(
                            "Error parsing TOML in {}".format(self.project_filename),
                            exception=e,
                        )
                    # convert toml to previous project structure:
                    new_definition = {}
                    if "rois" in self.definition.keys():
                        new_definition["rois"] = Project.__restructureProjectDefinition(
                            self.definition["rois"]
                        )
                    if "metrics" in self.definition.keys():
                        new_definition["metrics"] = (
                            Project.__restructureProjectDefinition(
                                self.definition["metrics"]
                            )
                        )
                    if "filters" in self.definition.keys():
                        new_definition["filters"] = (
                            Project.__restructureProjectDefinition(
                                self.definition["filters"]
                            )
                        )
                    self.definition = new_definition
                else:
                    logger.error("Unsupported project file type")
                    raise
        except FileNotFoundError as e:
            logger.error(f"File '{projectfilename}' not found.")
            raise e

        self.data = []

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.definition == other.definition
                and self.data == other.data
                and self.results == other.results
            )
        else:
            return False

    @staticmethod
    def __restructureProjectDefinition(def_dict: dict) -> list:
        new_def = []
        for k, v in def_dict.items():
            v["name"] = k
            new_def.append(v)
        return new_def

    def __load_single_datfile(self, filename: Path) -> pydre.core.DriveData:
        """Load a single .dat file (space delimited csv) into a DriveData object"""
        d = pl.read_csv(
            filename,
            separator=" ",
            null_values=".",
            truncate_ragged_lines=True,
            infer_schema_length=5000,
        )
        datafile_re_format0 = re.compile(
            "([^_]+)_Sub_(\\d+)_Drive_(\\d+)(?:.*).dat"
        )  # old format
        datafile_re_format1 = re.compile(
            "([^_]+)_([^_]+)_([^_]+)_(\\d+)(?:.*).dat"
        )  # [mode]_[participant id]_[scenario name]_[uniquenumber].dat
        match_format0 = datafile_re_format0.search(str(filename))

        if match_format0:
            experiment_name, subject_id, drive_id = match_format0.groups()
            drive_id = int(drive_id) if drive_id and drive_id.isdecimal() else None
            return pydre.core.DriveData.initV2(d, filename, subject_id, drive_id)
        elif match_format1 := datafile_re_format1.search(filename.name):
            mode, subject_id, scen_name, unique_id = match_format1.groups()
            return pydre.core.DriveData.initV4(
                d, filename, subject_id, unique_id, scen_name, mode
            )
        else:
            logger.warning(
                "Drivedata filename {} does not an expected format.".format(filename)
            )
            return pydre.core.DriveData(d, filename)

    # testing

    def processROI(
        self, roi: dict, datafile: pydre.core.DriveData
    ) -> list[pydre.core.DriveData]:
        """
        Handles running region of interest definitions for a dataset

        Args:
                roi: A dict containing the type of a roi and the filename of the data used to process it
                datafile: drive data object to process with the roi

        Returns:
                A list of drivedata objects containing the data for each region of interest
        """
        roi_type = roi["type"]
        if roi_type == "time":
            logger.info("Processing time ROI " + roi["filename"])
            roi_obj = pydre.rois.TimeROI(roi["filename"])
        elif roi_type == "rect":
            logger.info("Processing space ROI " + roi["filename"])
            roi_obj = pydre.rois.SpaceROI(roi["filename"])
        elif roi_type == "column":
            logger.info("Processing column ROI " + roi["columnname"])
            roi_obj = pydre.rois.ColumnROI(roi["columnname"])
        else:
            logger.warning("Unknown ROI type {}".format(roi_type))
            return [datafile]
        return roi_obj.split(datafile)

    def processFilter(
        self, datafilter: dict, datafile: pydre.core.DriveData
    ) -> pydre.core.DriveData:
        """
        Handles running any filter definition

        Args:
            datafilter: A dict containing the function of a filter and the parameters to process it

        Returns:
            The augmented DriveData object
        """
        ldatafilter = datafilter.copy()
        try:
            func_name = ldatafilter.pop("function")
            filter_func = pydre.filters.filtersList[func_name]
            datafilter_name = ldatafilter.pop("name")
        except KeyError as e:
            logger.error(
                'Filter definitions require a "function". Malformed filters definition: missing '
                + str(e)
            )
            raise e

        return filter_func(datafile, **ldatafilter)

    def processMetric(self, metric: dict, dataset: pydre.core.DriveData) -> dict:
        """

        :param metric:
        :param dataset:
        :return:
        """

        metric = metric.copy()
        try:
            func_name = metric.pop("function")
            metric_func = pydre.metrics.metricsList[func_name]
            report_name = metric.pop("name")
            col_names = pydre.metrics.metricsColNames[func_name]
        except KeyError as e:
            logger.warning(
                'Metric definitions require both "name" and "function". Malformed metrics definition'
            )
            raise e

        metric_dict = dict()
        if len(col_names) > 1:
            x = [metric_func(dataset, **metric)]
            report = pl.DataFrame(x, schema=col_names, orient="row")
            for i in range(len(col_names)):
                name = col_names[i - 1]
                data = x[0][i]
                metric_dict[name] = data
        else:
            # report = pl.DataFrame(
            #    [metric_func(dataset, **metric) ], schema=[report_name, ])
            metric_dict[report_name] = metric_func(dataset, **metric)
        return metric_dict

    # remove any parenthesis, quote mark and un-necessary directory names from a str
    def __clean(self, string):
        return string.replace("[", "").replace("]", "").replace("'", "").split("\\")[-1]

    def processDatafiles(
        self, datafilenames: list[Path], numThreads: int = 12
    ) -> Optional[pl.DataFrame]:
        """Load all metrics, then iterate over each file and process the filters, rois, and metrics for each.

        Args:
                datafilenames: a list of filename strings (SimObserver .dat files)
                numThreads: number of threads to run simultaneously in the thread pool

        Returns:
            metrics data for all metrics, or None on error

        """
        if "metrics" not in self.definition:
            logger.critical("No metrics in project file. No results will be generated")
            return None
        self.raw_data = []
        result_dataframe = pl.DataFrame()
        results_list = []
        # for datafilename in tqdm(datafilenames, desc="Loading files"):

        # with concurrent.futures.ThreadPoolExecutor(max_workers=numThreads) as executor:
        #    for result in executor.map(self.processSingleFile, datafilenames):
        #        for result_dict in result:
        #            results_list.append(result_dict)
        with tqdm(total=len(datafilenames)) as pbar:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=numThreads
            ) as executor:
                futures = {
                    executor.submit(self.processSingleFile, singleFile): singleFile
                    for singleFile in datafilenames
                }
                results = {}
                for future in concurrent.futures.as_completed(futures):
                    arg = futures[future]
                    try:
                        results[arg] = future.result()
                    except Exception as exc:
                        executor.shutdown(cancel_futures=True)
                        logger.critical("Unhandled Exception {}".format(exc))
                        sys.exit(1)

                    results_list.extend(future.result())
                    pbar.update(1)
        result_dataframe = pl.from_dicts(results_list)

        result_dataframe = result_dataframe.with_columns(
            pl.col("Subject").cast(pl.String),
            pl.col("ScenarioName").cast(pl.String),
            pl.col("ROI").cast(pl.String),
        )

        # Would just use try/except but polars throws overly alarming PanicException
        sorting_columns = ["Subject", "ScenarioName", "ROI"]
        # sorting_columns = [col for col in sorting_columns if col in result_dataframe.dtypes ]

        try:
            result_dataframe = result_dataframe.sort(sorting_columns)
        except pl.exceptions.PanicException as e:
            logger.warning("Can't sort results, must be missing a column.")

        self.results = result_dataframe
        return result_dataframe

    def processSingleFile(self, datafilename: Path):
        logger.info("Loading file #{}: {}".format(len(self.raw_data), datafilename))
        datafile = self.__load_single_datfile(datafilename)
        roi_datalist = []
        results_list = []

        if "filters" in self.definition:
            for datafilter in self.definition["filters"]:
                try:
                    datafile = self.processFilter(datafilter, datafile)
                except Exception as e:
                    logger.exception(
                        "Unhandled exception in {} while processing {}.".format(
                            datafilter, datafilename
                        )
                    )
                    raise e
        if "rois" in self.definition:
            for roi in self.definition["rois"]:
                try:
                    roi_datalist.extend(self.processROI(roi, datafile))
                except Exception as e:
                    logger.exception(
                        "Unhandled exception in {} while processing {}.".format(
                            roi, datafilename
                        )
                    )
                    raise e

        else:
            # no ROIs to process, but that's OK
            logger.warning("No ROIs, processing raw data.")
            roi_datalist.append(datafile)

        # if len(roi_datalist) == 0:
        # logger.warning("No ROIs found in {}".format(datafilename))
        roi_processed_metrics = []
        for data in roi_datalist:
            result_dict = {"Subject": data.PartID}
            if (
                data.format_identifier == 2
            ):  # these drivedata object was created from an old format data file
                result_dict["DriveID"] = datafile.DriveID
            elif (
                data.format_identifier == 4
            ):  # these drivedata object was created from a new format data file ([mode]_[participant id]_[scenario name]_[uniquenumber].dat)
                result_dict["Mode"] = self.__clean(str(data.mode))
                result_dict["ScenarioName"] = self.__clean(str(data.scenarioName))
                result_dict["UniqueID"] = self.__clean(str(data.UniqueID))
            result_dict["ROI"] = data.roi

            for metric in self.definition["metrics"]:
                try:
                    processed_metric = self.processMetric(metric, data)
                    result_dict.update(processed_metric)
                except Exception as e:
                    logger.critical(
                        "Unhandled exception {} in {} while processing {}.".format(
                            e.args, metric, datafilename
                        )
                    )
                    raise e
            results_list.append(result_dict)
        return results_list

    def saveResults(self, outfilename: pathlib.Path):
        """
        Args:
            outfilename: filename to output csv data to.

            The filename specified will be overwritten automatically.
        """
        try:
            self.results.write_csv(outfilename)
        except AttributeError:
            logger.error("Results not computed yet")



from typing import Optional

import re
import polars as pl
from polars import exceptions
import pydre.core
from pydre.core import ColumnsMatchError
from pydre.metrics import registerMetric
from loguru import logger
import numpy as np

from scipy import signal

# metrics defined here take a list of DriveData objects and return a single floating point value

# not registered & incomplete

# @registerMetric()
# def findFirstTimeAboveVel(drivedata: pydre.core.DriveData, cutoff: float = 25):
#     required_col = ["Velocity"]
#     drivedata.checkColumns(required_col)
#     timestepID = -1
#     # TODO: reimplement with selection and .head(1)
#
#     for i, row in drivedata.data.iterrows():
#         if row.Velocity >= cutoff:
#             timestepID = i
#             break
#
#     return timestepID


@registerMetric()
def colMean(
    drivedata: pydre.core.DriveData, var: str, cutoff: Optional[float] = None
) -> Optional[float]:
    """Calculates the mean of the specified column

    If `cutoff` is not `None`, then all values less than `cutoff` are ignored.
    If column is not numeric, `None` is returned.

    Parameters:
        var: The column name to process. Must be numeric.
        cutoff: Lower bound on data processed.

    Returns:
        Mean of selected column.
            If `cutoff` is not `None`, then all values less than `cutoff` are ignored.
            If column is not numeric, `None` is returned.
    """
    try:
        drivedata.checkColumnsNumeric([var])
    except ColumnsMatchError:
        return None
    if cutoff is not None:
        return (
            drivedata.data.get_column(var)
            .filter(drivedata.data.get_column(var) >= cutoff)
            .mean()
        )
    else:
        return drivedata.data.get_column(var).mean()


@registerMetric()
def colMedian(
    drivedata: pydre.core.DriveData, var: str, cutoff: Optional[float] = None
) -> Optional[float]:
    """Calculates the median of the specified column

    If `cutoff` is not `None`, then all values less than `cutoff` are ignored.
    If column is not numeric, `None` is returned.

    Parameters:
        var: The column name to process. Must be numeric.
        cutoff: Lower bound on data processed.

    Returns:
        Median of selected column.
            If `cutoff` is not `None`, then all values less than `cutoff` are ignored.
            If column is not numeric, `None` is returned.
    """
    try:
        drivedata.checkColumnsNumeric([var])
    except ColumnsMatchError:
        return None
    if cutoff is not None:
        return (
            drivedata.data.get_column(var)
            .filter(drivedata.data.get_column(var) >= cutoff)
            .median()
        )
    else:
        return drivedata.data.get_column(var).median()


@registerMetric()
def colSD(
    drivedata: pydre.core.DriveData, var: str, cutoff: Optional[float] = None
) -> Optional[float]:
    """Calculates the standard deviation of the specified column

    Uses Bessel's correction (denominator of N-1) for SD calculation.

    Parameters:
        var: The column name to process. Must be numeric.
        cutoff: Lower bound on data processed.

    Returns:
        Standard deviation of selected column.
            If `cutoff` is not `None`, then all values less than `cutoff` are ignored.
            If column is not numeric, `None` is returned.
    """
    try:
        drivedata.checkColumnsNumeric([var])
    except ColumnsMatchError:
        return None
    if cutoff is not None:
        return (
            drivedata.data.get_column(var)
            .filter(drivedata.data.get_column(var) >= cutoff)
            .std()
        )
    else:
        return drivedata.data.get_column(var).std()


@registerMetric()
def colMax(drivedata: pydre.core.DriveData, var: str) -> Optional[float]:
    """Calculates the maximum of the specified column

    Parameters:
        var: The column name to process. Must be numeric.

    Returns:
        Maximum value of selected column.
    """
    try:
        drivedata.checkColumnsNumeric([var])
    except ColumnsMatchError:
        return None
    return drivedata.data.get_column(var).max()


@registerMetric()
def colMin(drivedata: pydre.core.DriveData, var: str) -> Optional[float]:
    """Calculates the minimum of the specified column

    Parameters:
        var: The column name to process. Must be numeric.

    Returns:
        Minimum value of selected column.
    """
    try:
        drivedata.checkColumnsNumeric([var])
    except ColumnsMatchError:
        return None
    return drivedata.data.get_column(var).min()


@registerMetric()
def colFirst(drivedata: pydre.core.DriveData, var: str) -> Optional[float]:
    """Returns the first value of the specified column

    Parameters:
        var: The column name to process. Must be numeric.

    Returns:
        First value of selected column.
    """

    try:
        drivedata.checkColumnsNumeric([var])
    except ColumnsMatchError:
        return None
    return drivedata.data.get_column(var).head(1).item()


@registerMetric()
def colLast(drivedata: pydre.core.DriveData, var: str) -> Optional[float]:
    """Returns the last value of the specified column

    Parameters:
        var: The column name to process. Must be numeric.

    Returns:
        Last value of selected column.
    """
    try:
        drivedata.checkColumnsNumeric([var])
    except ColumnsMatchError:
        return None
    return drivedata.data.get_column(var).tail(1).item()


@registerMetric()
def timeAboveSpeed(
    drivedata: pydre.core.DriveData, cutoff: float = 0, percentage: bool = False
) -> Optional[float]:
    """Returns the amount of seconds travelling above the specified speed

    Parameters:
        cutoff: Speed threshold value
        percentage: If true, return the percentage of total time in the DriveData object where the speed was
            above the cutoff.

    Note: Requires data columns
        - SimTime: Time in seconds
        - Velocity: Speed

    Returns:
        Number of seconds travelling above the specified speed.
    """
    required_col = ["SimTime", "Velocity"]
    # to verify if column is numeric
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None
    drivedata.checkColumns(required_col)

    df = drivedata.data.select(
        [
            pl.col("SimTime"),
            pl.col("Velocity"),
            pl.col("SimTime").diff().clip(lower_bound=0).alias("Duration"),
        ]
    )

    time = df.get_column("Duration").filter(df.get_column("Velocity") >= cutoff).sum()
    try:
        total_time = df.get_column("SimTime").max() - df.get_column("SimTime").min()
    except TypeError:
        return None
    if percentage:
        out = time / total_time
    else:
        out = time
    return out


@registerMetric()
def timeWithinSpeedLimit(
    drivedata: pydre.core.DriveData, lowerlimit: float = 0, percentage: bool = False
) -> Optional[float]:
    """Returns the amount of seconds travelling below the speed limit

    Parameters:
        lowerlimit: Filter out all velocities lower than this value.
        percentage: If true, return the percentage of total time in the DriveData object where the speed was under the limit

    Note: Requires data columns
        - SimTime: Time in seconds
        - Velocity: Speed in meters per second
        - SpeedLimit: Speed limit in miles per hour

    Returns:
        Number of seconds travelling above the specified speed.
    """
    required_col = ["SimTime", "Velocity", "SpeedLimit"]
    # to verify if column is numeric
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None

    df = drivedata.data.select(
        [
            pl.col("SimTime"),
            pl.col("Velocity"),
            pl.col("Velocity").mul(2.23694).alias("VelocityMPH"),
            pl.col("SimTime").diff().clip(lower_bound=0).alias("Duration"),
            pl.col("SpeedLimit"),
        ]
    )

    time = (
        df.filter(
            (pl.col("VelocityMPH") <= pl.col("SpeedLimit"))
            & (pl.col("VelocityMPH") >= lowerlimit)
        )
        .get_column("Duration")
        .sum()
    )

    if percentage:
        total_time = df.get_column("SimTime").max() - df.get_column("SimTime").min()
        output = time / total_time
    else:
        output = time
    return output


@registerMetric()
def stoppingDist(
    drivedata: pydre.core.DriveData, roadtravelposition="XPos"
) -> Optional[float]:
    """Returns the position of the first complete stop of the vehicle, relative to the center position of the
    DriveData object, as measured by the `roadtravelposition` column.

    This metric is designed to be used with the position-based ROI to define ROIs that have the stop line for a
    stop sign or stop light as the middle value in the direction of travel.

    Parameters:
        roadtravelposition: column name of increasing distance of travel, in meters

    Note: Requires data columns
        - Velocity: Speed in meters per second

    Returns:
        Number of meters the detected stop is away from the stop line. If no stops detected, return `10000`
    """
    required_col = [roadtravelposition, "Velocity"]
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None

    df = drivedata.data.select([pl.col(roadtravelposition), pl.col("Velocity")])
    if df.height == 0:
        return None
    # filtering the data to remove any negative velocities and only focus on velocities below 0.01 (chosen epsilon)
    velocities = df.filter(
        (df.get_column("Velocity") >= 0.0) & (df.get_column("Velocity") < 0.01)
    )

    lineposition = (
        df.get_column(roadtravelposition).min()
        + df.get_column(roadtravelposition).max()
    ) / 2

    if velocities.height == 0:
        # hard coding a "bad" value for stopposition if vehicle never stopped
        return 10000
    else:
        # finding the furthest position where velocity was close to zero
        stopposition = velocities.item(velocities.height - 1, 0)
    return lineposition - stopposition


@registerMetric()
def maxdeceleration(
    drivedata: pydre.core.DriveData, cutofflimit: float = 1
) -> Optional[float]:
    """Returns the maximum deceleration value

    Parameters:
        cutofflimit: all rows with velocities less than this value will be removed

    Note: Requires data columns
        - LonAccel: Acceleration value in the direction of travel, in meters per second per second
        - Velocity: Speed in meters per second

    Returns:
        maximum negative acceleration value (i.e. minimum acceleration value)
    """
    required_col = ["LonAccel", "Velocity"]
    # to verify if column is numeric
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None

    df = drivedata.data.select([pl.col("LonAccel"), pl.col("Velocity")])

    # filtering data to take out all rows with velocities less than cutoff value
    dfupdated = df.filter(df.get_column("Velocity") > cutofflimit)
    decel = dfupdated.filter(dfupdated.get_column("LonAccel") < 0)

    maxdecel = decel.get_column("LonAccel").min()
    return maxdecel


@registerMetric()
def maxacceleration(
    drivedata: pydre.core.DriveData, cutofflimit: int = 1
) -> Optional[float]:
    """Returns the maximum acceleration value

    Parameters:
        cutofflimit: all rows with velocities less than this value will be removed

    Note: Requires data columns
        - LonAccel: Acceleration value in the direction of travel, in meters per second per second
        - Velocity: Speed in meters per second

    Returns:
        maximum acceleration value
    """
    required_col = ["LonAccel", "Velocity"]
    # to verify if column is numeric
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None

    df = drivedata.data.select([pl.col("LonAccel"), pl.col("Velocity")])
    dfupdated = df.filter(df.get_column("Velocity") > cutofflimit)
    accel = dfupdated.filter(dfupdated.get_column("LonAccel") > 0)
    maxaccel = accel.filter(
        accel.get_column("LonAccel") == accel.get_column("LonAccel").max()
    )

    return maxaccel


@registerMetric()
def numbrakes(
    drivedata: pydre.core.DriveData, cutofflimit: float = 1
) -> Optional[float]:
    """Returns the number of times the brakes were pressed

    Parameters:
        cutofflimit: minimum velocity at which braking should be considered

    Note: Requires data columns
        - BrakeStatus: Column indicating braking. 0: no braking, greater than 0: braking
        - Velocity: Speed

    Returns:
        number of separate braking events
    """
    required_col = ["Brake", "Velocity"]
    # to verify if column is numeric
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None

    df = drivedata.data.select(
        [
            pl.col("Brake"),
            pl.col("Brake").gt(0).cast(pl.Float32).diff().alias("BrakeStatus"),
            pl.col("Velocity"),
        ]
    )
    dfupdated = df.filter(df.get_column("Velocity") > cutofflimit)

    numberofbrakes = (
        dfupdated.filter(dfupdated.get_column("BrakeStatus") > 0)
        .select(pl.count())
        .item(0, 0)
    )

    return numberofbrakes


def _calculateReversals(df):
    k = 1
    n = 0
    threshold = 0.0523598776 * 2
    for l in range(2, len(df)):
        if df[l] - df[k] >= threshold:
            n = n + 1
            k = l
        elif df[l] <= df[k]:
            k = l
    return n


@registerMetric()
def steeringReversalRate(drivedata: pydre.core.DriveData) -> float:
    """Steering reversal rate

    As defined in [SAE j2944](https://www.auto-ui.org/docs/sae_J2944_appendices_PG_130212.pdf)

    Note: Requires data columns
        - SimTime: simulation time
        - Steer: orientation of steering wheel in radians

    Returns:
        reversals per minute

    """
    #
    required_col = ["SimTime", "Steer"]
    # to verify if column is numeric
    drivedata.checkColumnsNumeric(required_col)
    drivedata.checkColumns(required_col)

    df = drivedata.data.select([pl.col("SimTime"), pl.col("Steer")])
    # convert to numpy and resample data to have even time steps (of 32Hz)
    df = df.slice(1, None)
    numpy_df = df.to_numpy()
    original_time = numpy_df[:, 0]
    original_steer = numpy_df[:, 1]

    # 32Hz = 0.03125seconds
    new_time = np.arange(original_time[0], original_time[-1], 0.03125)
    new_steer = np.interp(new_time, original_time, original_steer)
    numpy_df = np.column_stack((new_time, new_steer))

    # apply second order butterworth filter at 6z frequency
    sos = signal.butter(2, 6, output="sos", fs=32)
    theta_i = signal.sosfilt(sos, numpy_df[:, 1], zi=None)

    # calcuate thetaI
    theta_prime_i = np.diff(theta_i)
    theta_prime_i = np.append(0, theta_prime_i)

    # make column for i and combine with theta_prime_i column
    i = np.arange(1, len(theta_prime_i) + 1)
    df_with_theta = np.column_stack((i, theta_prime_i, theta_i))

    # find all values of i where theta_prime_i is 0
    df = pl.from_numpy(
        df_with_theta, schema=["i", "thetaPrimeI", "thetaI"], orient="row"
    )
    df_except_one = df.filter(df.get_column("i") > 1)
    zeros = df_except_one.filter(df_except_one.get_column("thetaPrimeI") == 0).drop(
        "thetaPrimeI"
    )

    # find values of i where difference between consequential signs of theta_prime_i is 2
    sign_diff = df.with_columns(
        df.get_column("thetaPrimeI").sign().diff(-1).alias("SignDiff")
    )
    diff_two = sign_diff.filter(
        (sign_diff.get_column("SignDiff") == 2)
        | (sign_diff.get_column("SignDiff") == -2)
    ).drop(["SignDiff", "thetaPrimeI"])

    # merge list of i's to get one sorted list of i's
    set_of_i = zeros.merge_sorted(diff_two, key="i").to_numpy()

    # calculate total reversals
    n_upwards = _calculateReversals(set_of_i[:, 1])
    set_of_i_down = np.multiply(set_of_i[:, 1], -1)
    n_downwards = _calculateReversals(set_of_i_down)
    reversals = n_upwards + n_downwards

    # reversal rate as reversals/ minute
    reversal_rate = reversals / ((np.max(original_time) - np.min(original_time)) / 60)
    return reversal_rate


@registerMetric()
def throttleReactionTime(drivedata: pydre.core.DriveData) -> Optional[float]:
    """ Calculates the time it takes to accelerate once follow car brakes (r2d)

        Note: Requires data columns
            - SimTime: Simulation time in secondsSimTime: simulation time
            - FollowCarBraking Status: Whether the follow car is braking
            - LonAccel: Longitude acceleration
            - Brake: Whether the ownship is braking

        Returns:
            Time in seconds from when the follow car braked to when the ownship started accelerating forward.
    """

    required_col = ["FollowCarBrakingStatus", "LonAccel", "SimTime", "Brake"]

    try:
        drivedata.checkColumnsNumeric(required_col)
    except pl.exceptions.PolarsError:
        return None

    df = drivedata.data.select(
        [
            pl.col("FollowCarBrakingStatus"),
            pl.col("SimTime"),
            pl.col("LonAccel"),
            pl.col("Brake"),
            pl.col("Throttle"),
        ]
    )

    if df.height < 1:
        return None

    initial_time = df.get_column("SimTime").item(0)

    try:
        df = df.filter(
            pl.col("SimTime")
            > df.filter(pl.col("Brake") > 3.0).get_column("SimTime").item(0)
        )
    except IndexError:
        logger.warning(
            f"No braking detected for roi {drivedata.roi} in file {drivedata.sourcefilename}"
        )
        return None

    df_after_brake = df.filter(pl.col("Brake") == 0)

    try:
        time_of_accel = (
            df_after_brake.filter(pl.col("LonAccel") > 0).get_column("SimTime").item(0)
        )
    except IndexError:
        logger.warning(
            f"No subsequent acceleration detected for roi {drivedata.roi} in file {drivedata.sourcefilename}"
        )
        return None

    throttle_reaction_time = time_of_accel - initial_time
    return throttle_reaction_time

@registerMetric()
def maxAcceleration(drivedata: pydre.core.DriveData) -> Optional[float]:
    required_col = ["LatAccel", "LonAccel", "SimTime"]

    drivedata.checkColumnsNumeric(required_col)

    df = drivedata.data.select([pl.col("LatAccel"), pl.col("LonAccel"), pl.col("SimTime")])

    df = df.with_columns(((pl.col("LonAccel") ** 2) + (pl.col("LatAccel") ** 2)).sqrt().alias("Total_Accel"))
    return df.get_column("Total_Accel").max()




# laneExits
# Will compute the number of transitions from the lane number specified to (lane+1) or (lane-1)
# the 'sign' function will remove transitions from (lane+1) to (lane+2) or similar
@registerMetric()
def laneExits(drivedata: pydre.core.DriveData, lane=2, lane_column="Lane"):
    drivedata.checkColumnsNumeric([lane_column])
    return (
        drivedata.data.select((pl.col(lane_column) - lane).sign().diff().abs())
        .sum()
        .item()
    )


def laneViolations(
    drivedata: pydre.core.DriveData,
    offset: str = "LaneOffset",
    lane: int = 2,
    lane_column: str = "Lane",
    lane_width: float = 3.65,
    car_width: float = 2.1,
):
    drivedata.checkColumnsNumeric([lane_column])
    # tolerance is the maximum allowable offset deviation from 0
    tolerance = lane_width / 2 - car_width / 2
    lane_data = drivedata.data.filter(pl.col(lane_column) == 2)
    lane_data = lane_data.with_columns(
        pl.col(offset).abs().is_between(upper_bound=tolerance)
    )

    is_violating = lane_data.get_column(offset).abs() > tolerance
    return is_violating.diff().clip_min(0).sum()


def laneViolationDuration(
    drivedata: pydre.core.DriveData,
    offset: str = "LaneOffset",
    lane: int = 2,
    lane_column: str = "Lane",
    lane_width: float = 3.65,
    car_width: float = 2.1,
):
    required_col = [offset, lane_column, "LaneDuration", "SimTime", "Duration"]
    # to verify if column is numeric
    drivedata.checkColumnsNumeric(required_col)

    tolerance = lane_width / 2 - car_width / 2
    lane_data = drivedata.data.filter(pl.col(lane_column) == 2)
    lane_data = lane_data.select(
        [
            pl.col(offset).abs().alias("LaneViolations") > tolerance,
            pl.col("SimTime").diff().clip_min(0).alias("Duration"),
        ]
    )

    return lane_data.select(pl.col("LaneViolations") * pl.col("Duration")).sum().item()


# Parameters:

# Offset: the name of the specific column in the datafile, could be LaneOffset or RoadOffset

# noisy: If this is set to 'true', a low pass filter with 5 Hz cut off frequency will be applied, according to documentation
# The document doesn't specify the order of filter so I'll use 1st order here

# filfilt: if this is set to 'true', the filter will be applied twice, once forward and once backwards. This gives better results
# when testing with a sin signal, but I'm not sure if that leads to a risk or not so I'll keep that as an option

# noisy and filtfilt are NOT case sensitive and are meaningful only when calculating MSDLP
# @registerMetric()
# def lanePosition(drivedata: pydre.core.DriveData, laneInfo: str = "sdlp", lane=2, lane_width: float = 3.65,
#                  car_width: float = 2.1,
#                  offset: str = "LaneOffset", noisy="false", filtfilt="false"):
#     required_col = ["SimTime", "DatTime", "Lane", offset]
#     drivedata.checkColumns(required_col)
#
#     df = pandas.DataFrame(drivedata.data, columns=required_col)  # drop other columns
#     LPout = None
#     if (df.size > 0):
#         if (laneInfo in ["mean", "Mean"]):
#             # mean lane position
#             LPout = np.mean((df[offset]))  # abs to give mean lane "error"
#         elif (laneInfo in ["msdlp", "MSDLP"]):
#             samplingFrequency = 1 / np.mean(np.diff(df.DatTime))  # calculate sampling drequency based on DatTime
#             # samplingFrequency = 1 / np.mean(np.diff(df.SimTime))
#
#             sos = signal.butter(2, 0.1, 'high', analog=False, output='sos',
#                                 fs=float(samplingFrequency))  # define butterWorthFilter
#             # the output parameter can also be set to 'ba'. Under this case, signal.lfilter(b, a, array) or
#             # signal.filtfilt(b, a, array) should be used. sos is recommanded for general purpose filtering
#
#             data = df[offset]
#             if (noisy.lower() == "true"):
#                 sosLow = signal.butter(1, 5, 'low', analog=False, output='sos', fs=float(samplingFrequency))
#                 data = signal.sosfilt(sosLow, data)
#                 # apply a low pass filter to reduce the noise
#
#             filteredLP = None
#             if (filtfilt.lower() == "true"):
#                 filteredLP = signal.sosfiltfilt(sos, data)  # apply the filter twice
#             else:
#                 filteredLP = signal.sosfilt(sos, data)  # apply the filter once
#             # signal.sosfiltfilt() applies the filter twice (forward & backward) while signal.sosfilt applies
#             # the filter once.
#
#             LPout = np.std(filteredLP)
#
#         elif (laneInfo in ["sdlp", "SDLP"]):
#             LPout = np.std(df[offset])
#             # Just cause I've been staring at this a while and want to get some code down:
#             # Explanation behind this: SAE recommends using the unbiased estimator "1/n-1". The np code does
#             # not use this, so I wrote up code that can easily be subbed in, if it's determined necessary.
#             """
#             entrynum = len(df[offset])
#             unbiased_estimator = 1/(entrynum - 1)
#             average = np.mean((df[offset]))
#             variation = 0
#             for entry in entrynum:
#                 variation += (pow(df[offset][entry] - average, 2))
#             LPout = math.sqrt(unbiased_estimator * variation)
#             """
#         elif (laneInfo in ["exits"]):
#             LPout = (df.Lane - df.Lane.shift(1)).abs().sum()
#         elif (laneInfo in ["violation_count"]):
#             LPout = 0
#             # tolerance is the maximum allowable offset deviation from 0
#             tolerance = lane_width / 2 - car_width / 2
#             is_violating = abs(df[offset]) > tolerance
#
#             # Shift the is_violating array and look for differences.
#             shifted = is_violating.shift(1)
#             shifted.iloc[0] = is_violating.iloc[0]
#
#             # Create an array which is true whenever the car goes in/out of
#             # the lane
#             compare = shifted != is_violating
#
#             # shiftout becomes an array which only has elements each time
#             # compare is true (ie, violation status changed). These elements
#             # are True when the direction is out of the lane, False when the
#             # direction is back into the lane. We only count the out shifts.
#             shifts = compare.loc[compare == True] == is_violating.loc[compare == True]
#             shiftout = shifts.loc[shifts == True]
#
#             # Count all violations. Add one if the region begins with a violation.
#             if is_violating.iloc[0] is True:
#                 LPout = LPout + 1
#             LPout = LPout + shiftout.size
#
#         elif laneInfo in ["violation_duration"]:
#             LPout = 0
#             tolerance = lane_width / 2 - car_width / 2
#             violations = df[abs(df[offset]) > tolerance]
#             if (violations.size > 0):
#                 deltas = violations.diff()
#                 deltas.iloc[0] = deltas.iloc[1]
#                 LPout = sum(deltas.SimTime[(deltas.SimTime < .5) & (deltas.SimTime > 0)])
#         else:
#             print("Not a valid lane position metric - use 'mean', 'sdlp', or 'exits'.")
#             return None
#     return LPout


@registerMetric()
def roadExits(drivedata: pydre.core.DriveData):
    required_col = ["SimTime", "RoadOffset", "Velocity"]
    # to verify if column is numeric
    drivedata.checkColumnsNumeric(required_col)
    drivedata.checkColumns(required_col)

    # assuming a two lane road, determine the amount of time they were not in the legal roadway
    # Lane width 3.6m, car width 1.8m

    df = drivedata.data.select(
        [
            pl.col("SimTime"),
            pl.col("RoadOffset"),
            pl.col("Velocity"),
            pl.col("SimTime")
            .diff()
            .alias("Duration")
            .clip(
                0, 0.5
            ),  # any duration outside this range is bad data (maybe from a splice)
        ]
    )

    outtimes = df.filter(
        pl.col("RoadOffset").is_between(0, 7.2).is_not() & (pl.col("Velocity") > 1)
    )

    return outtimes.get_column("Duration").sum()


@registerMetric()
def roadExitsY(drivedata: pydre.core.DriveData):
    required_col = ["SimTime", "YPos", "Velocity"]
    # to verify if column is numeric
    drivedata.checkColumnsNumeric(required_col)
    drivedata.checkColumns(required_col)

    # assuming a two lane road, determine the amount of time they were not in the legal roadway
    # Lane width 3.6m, car width 1.8m

    df = drivedata.data.select(
        [
            pl.col("SimTime"),
            pl.col("RoadOffset"),
            pl.col("Velocity"),
            pl.col("SimTime")
            .diff()
            .alias("Duration")
            .clip(
                0, 0.5
            ),  # any duration outside this range is bad data (maybe from a splice)
        ]
    )

    outtimes = df.filter(
        pl.col("RoadOffset").is_between(0 + 0.9, 7.2 - 0.9).is_not()
        & (pl.col("Velocity") > 1)
        # TODO: double check offsets
    )

    return outtimes.get_column("Duration").sum()


# cutoff doesn't work
@registerMetric()
def steeringEntropy(drivedata: pydre.core.DriveData, cutoff: float = 0):
    required_col = ["SimTime", "Steer"]
    # to verify if column is numeric
    drivedata.checkColumnsNumeric(required_col)
    drivedata.checkColumns(required_col)

    out = []
    df = drivedata.data.select([pl.col("SimTime"), pl.col("Steer")])
    df = df.unique(
        subset=["Steer"], maintain_order=True
    )  # remove nans and drop duplicates

    if df.select(pl.count()).item(0, 0) == 0:
        return None

    # downsample the array to change the time step to every 0.15 seconds
    # based on article "Development of a Steering Entropy Method For Evaluating Driver Workload"
    df = df.slice(1, None)
    numpy_df = df.to_numpy()
    original_time = numpy_df[:, 0]
    original_steer = numpy_df[:, 1]

    new_time = np.arange(original_time[0], original_time[-1], 0.15)
    new_steer = np.interp(new_time, original_time, original_steer)

    numpy_df = np.column_stack((new_time, new_steer))
    df = pl.from_numpy(numpy_df, schema=["SimTime", "Steer"], orient="row")

    # calculate predicted angle
    # pAngle(n) = angle(n-1) + (angle(n-1) - angle(n-2)) + 0.5*( (angle(n-1) - angle(n-2)) - (angle(n-2) - angle(n-3)))
    end_index = df.height - 1

    angle_minus_one = df.get_column("Steer").slice(2, end_index - 2).to_numpy()
    angle_minus_two = df.get_column("Steer").slice(1, end_index - 2).to_numpy()
    angle_minus_three = df.get_column("Steer").slice(0, end_index - 2).to_numpy()
    pAngle = (
        angle_minus_one
        + (angle_minus_one - angle_minus_two)
        + 0.5
        * ((angle_minus_one - angle_minus_two) - (angle_minus_two - angle_minus_three))
    )
    pAngle = pl.from_numpy(pAngle, schema=["Steer"], orient="row")

    # calculate error
    error = df.get_column("Steer").slice(3, None) - pAngle.get_column("Steer")

    # TIYA - stopped making edits to this function this point onwards
    out.append(error)

    # concatenate out (list of np objects) into a single list
    if len(out) == 0:
        return None
    error = np.concatenate(out)
    # 90th percentile of error
    alpha = np.nanpercentile(error, 90)

    # divide into 9 bins with edges: -5a,-2.5a,-a,a,2.5,5a
    binnedError = np.histogram(
        error,
        bins=[
            -10 * abs(alpha),
            -5 * abs(alpha),
            -2.5 * abs(alpha),
            -abs(alpha),
            -0.5 * abs(alpha),
            0.5 * abs(alpha),
            abs(alpha),
            2.5 * abs(alpha),
            5 * abs(alpha),
            10 * abs(alpha),
        ],
    )

    # calculate H
    binnedArr = np.asarray(binnedError[0])
    binnedArr = binnedArr.astype(float)

    # check out calc of p
    p = np.divide(binnedArr, np.sum(binnedArr))
    Hp = np.multiply(np.multiply(-1, p), (np.log(p) / np.log(9)))
    Hp = Hp[~np.isnan(Hp)]
    Hp = np.sum(Hp)

    return Hp


@registerMetric()
def closeFollowing(
    drivedata: pydre.core.DriveData,
    threshold: float = 2,
    percentage: bool = False,
    minvelocity: Optional[float] = None,
) -> Optional[float]:
    """Close following is the amount of time where the headway time was lower than a threshold

    Very close following is sometimes called tailgating

    Parameters:
        threshold: maximum following time to collision, in seconds
        percentage: flag to output a percentage of total time spent
            following closely, rather than the number of seconds
        minvelocity: minimum velocity to consider rows as part of total time.
            If None, consider all times

    Note: Requires data columns:
        - SimTime: Simulation time in seconds
        - HeadwayTime: Headway time in seconds. This is the headway distance multiplied by the current
            speed of the vehicle.

    Returns:
        number of seconds or percentage of close following, under the specified threshold

    """
    # to verify if column is numeric
    required_col = ["SimTime", "HeadwayTime"]
    if minvelocity:
        required_col.append("Velocity")

    try:
        drivedata.checkColumnsNumeric(required_col)
    except pl.exceptions.PolarsError:
        return None

    following_df = drivedata.data.with_columns(
        pl.col("SimTime").diff().alias("delta_t")
    )

    # find all tailgating instances where the delta time is reasonable.
    # this ensures we don't get screwy data from merged files
    following_df = following_df.filter(pl.col("delta_t").is_between(0, 0.5))

    if minvelocity:
        following_df = following_df.filter(pl.col("Velocity") >= minvelocity)

    tail_time = (
        following_df.filter(
            pl.col("HeadwayTime").is_between(0, threshold, closed="none")
        )
        .select("delta_t")
        .sum()
        .item()
    )

    total_time = following_df.select("delta_t").sum().item()

    if percentage:
        if total_time > 0:
            return tail_time / total_time
        else:
            return None
    else:
        return tail_time


# determines when the ownship collides with another vehicle by examining headway distance as threshold
@registerMetric()
def leadVehicleCollision(
    drivedata: pydre.core.DriveData, cutoff: float = 2.85
) -> Optional[float]:
    """Number of collisions between the ownship and the lead vehicle.

    Parameters:
        cutoff: number of meters below wich the headway distance is considered "collided".
            Since headway distance is measured center-to-center of the two vehicles, 2.85m is used
            as the default value based on the size of sedans in the simulation.

    Note: Requires data columns
        - SimTime: Simulation time in seconds
        - HeadwayDistance: distance between the ownship and the lead vehicle in meters

    Returns:
        number of collision events

    """
    required_col = ["SimTime", "HeadwayDistance"]
    # to verify if column is numeric
    try:
        drivedata.checkColumnsNumeric(required_col)
    except pl.exceptions.PolarsError:
        return None

    # find contiguous instances of headway distance < the cutoff
    collision_table = drivedata.data.select(
        (pl.col("HeadwayDistance") <= cutoff).alias("CollisionZone")
    )
    collisions = (
        collision_table.select(pl.col("CollisionZone").cast(pl.Int32).diff().clip(0, 1))
        .sum()
        .item()
    )
    return collisions


def _firstOccurrence(df: pl.DataFrame, column: str):
    try:
        output = df[column].head(1)
        return output.index[0]
    except pl.PolarsError:
        return None


@registerMetric()
def timeFirstTrue(
    drivedata: pydre.core.DriveData, var: str, timecol: str = "SimTime"
) -> Optional[float]:
    """Time of the first true (>0) value in the specified variable column

    Parameters:
        var: column to check for first true value
        timecol: column to use for for time

    Returns:
        earliest value of the time column when the var column is true (above 0). None is returned if
            the var column is never true or if there is a error in processing.

    """

    required_col = [var, timecol]
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None

    try:
        df = drivedata.data.filter(pl.col(var) > 0)
    except pl.exceptions.ComputeError as e:
        logger.warning(
            "Failure processing timeFirstTrue metric for variable {} in file {}".format(
                var, drivedata.sourcefilename
            )
        )
        return None
    if df.height == 0:
        return None
    return (
        df.select("SimTime").head(1).item()
        - drivedata.data.select("SimTime").head(1).item()
    )


@registerMetric()
def reactionBrakeFirstTrue(
    drivedata: pydre.core.DriveData, var: str
) -> Optional[float]:
    required_col = [var, "SimTime"]
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None

    try:
        df = drivedata.data.filter(pl.col(var) > 5)
    except pl.exceptions.ComputeError as e:
        logger.warning("Brake value non-numeric in {}".format(drivedata.sourcefilename))
        return None
    if drivedata.data.height == 0 or df.height == 0:
        return None
    return (
        df.select("SimTime").head(1).item()
        - drivedata.data.select("SimTime").head(1).item()
    )


@registerMetric()
def reactionTimeEventTrue(drivedata: pydre.core.DriveData, var1: str, var2: str):
    required_col = [var1, var2, "SimTime"]
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None
    break_reaction = reactionBrakeFirstTrue(drivedata, var1)

    if break_reaction:
        return break_reaction
    else:
        df = drivedata.data.filter(abs(pl.col(var2)) >= 0.2)
        if drivedata.data.height == 0 or df.height == 0:
            return None
        return (
            df.select("SimTime").head(1).item()
            - drivedata.data.select("SimTime").head(1).item()
        )


"""
TBI Reaction algorithm
To find the braking reaction time to the event (for each Section):
    Only look at values after the event E (Activation = 1)
    Find the first timestep X where:
		X is after the Event (Activation=1)
			AND
		BrakeForce(X) – BrakeForce(E) > 0
    Reaction time is Time(X) – Time(First timestep where Activation = 1)


To find the throttle reaction time to the event (for each Section):
    Only look at values after the event E (Activation = 1)
    Find the first timestep X where:
		X is after the Event (Activation=1)
			AND
		Throttle(X) – Throttle(X-1) > SD([Throttle for all X in Section where Activation !=1]).
    Reaction time is Time(X) – Time(First timestep where Activation = 1)


Then you average the two throttle reaction times and the two brake reaction times for each of the 4 DriveIDs.
This results in 8 reaction times per participant.
"""


@registerMetric()
def reactionTime(drivedata: pydre.core.DriveData, brake_cutoff=1, steer_cutoff=0.2):
    required_col = ["SimTime", "Brake", "Steer", "XPos", "HeadwayDistance"]
    # to verify if column is numeric
    try:
        drivedata.checkColumnsNumeric(required_col)
    except ColumnsMatchError:
        return None

    df = drivedata.data.select(
        [
            pl.col("SimTime"),
            pl.col("Steer"),
            pl.col("Brake"),
            pl.col("XPos"),
            pl.col("HeadwayDistance"),
        ]
    )

    event_start_time = df.get_column("SimTime").item(0)
    # calculate braking reaction time
    brake_df = df.filter(df.get_column("Brake") > brake_cutoff)
    if brake_df.is_empty():
        brake_reaction = 50
    else:
        first_brake = brake_df.get_column("SimTime").item(0)
        brake_reaction = first_brake - event_start_time
    # calculate swerving reaction time
    first_steer = df.get_column("Steer").item(0)
    try:
        df_steer = df.with_columns(
            (pl.col("Steer") - first_steer).alias("SteerDiff").abs()
        )
        df_steer = df_steer.filter(df_steer.get_column("SteerDiff") > steer_cutoff)
    except exceptions.InvalidOperationError:
        return None

    if df_steer.is_empty():
        steer_reaction = 50
    else:
        steer_reaction = df_steer.get_column("SimTime").item(0) - event_start_time

    if steer_reaction > 10 and brake_reaction > 10:
        reactionTime = None
    else:
        reactionTime = min(steer_reaction, brake_reaction)
    return reactionTime


@registerMetric()
def criticalEventStartPos(drivedata: pydre.core.DriveData):
    required_col = ["XPos"]
    # to verify if column is numeric
    drivedata.checkColumnsNumeric(required_col)
    drivedata.checkColumns(required_col)

    df = drivedata.data.select(pl.col("XPos"))
    return df.get_column("XPos").item(0)


@registerMetric()
def criticalEventEndPos(drivedata: pydre.core.DriveData):
    required_col = ["XPos"]
    # to verify if column is numeric
    drivedata.checkColumnsNumeric(required_col)
    drivedata.checkColumns(required_col)

    df = drivedata.data.select(pl.col("XPos"))
    return df.get_column("XPos").item(-1)


@registerMetric(
    "R2DIDColumns", ["ParticipantID", "MatchID", "Case", "Location", "Gender", "Week"]
)
def R2DIDColumns(drivedata: pydre.core.DriveData):
    ident = drivedata.PartID
    ident_groups = re.match(r"(\d)(\d)(\d)(\d\d\d\d)[wW](\d)", ident)
    if ident_groups is None:
        logger.warning("Could not parse R2D ID " + ident)
        return [None, None, None, None, None, None]
    participant_id = (
        ident_groups.group(1)
        + ident_groups.group(2)
        + ident_groups.group(3)
        + ident_groups.group(4)
    )
    case = ident_groups.group(1)
    if case == "3":
        case = "Case"
    elif case == "5":
        case = "Control"
    elif case == "7":
        case = "Enrolled"
    location = ident_groups.group(2)
    if location == "1":
        location = "UAB"
    elif location == "2":
        location = "OSU"
    elif location == "3":
        location = "UA"
    gender = ident_groups.group(3)
    if gender == "1":
        gender = "Male"
    elif gender == "2":
        gender = "Female"
    match_id = ident_groups.group(4)
    week = ident_groups.group(5)
    return week, participant_id, match_id, case, location, gender



import pathlib
from abc import ABCMeta, abstractmethod

import pydre.core
import polars as pl
import re
from loguru import logger
from collections.abc import Iterable


class ROIProcessor(object, metaclass=ABCMeta):
    roi_list = []

    @abstractmethod
    def __init__(self, filename: str | pathlib.Path, nameprefix: str = ""):
        pass

    @abstractmethod
    def split(
        self, sourcedrivedata: pydre.core.DriveData
    ) -> list[pydre.core.DriveData]:
        """Splits the drivedata object according to the ROI specifications.

        Parameters:
            sourcedrivedata: input drivedata object

        Returns:
            list of drivedata objects after splitting
        """
        pass


def sliceByTime(
    begin: float, end: float, column: str, drive_data: pl.DataFrame
) -> pl.DataFrame:
    """
        args:
            begin: float defnining the start point of the slice
            end: float defining the end part of the slice
            column: which column in the drive_data frame to use for the time.  This is usually SimTime or VidTime.
            drive_data: polars DataFrame containing the data to be sliced

        returns:
            polars.DataFrame slice containing requested time slice

    Given a start and end and a column name that represents a time value, output the slice that contains
    only the specified data.
    """
    try:
        dataframeslice = drive_data.filter(
            pl.col(column).is_between(begin, end, closed="left")
        )
    except KeyError:
        logger.error("Problem in applying Time ROI to using time column " + column)
        dataframeslice = drive_data
    return dataframeslice


class TimeROI(ROIProcessor):
    def __init__(self, filename: str | pathlib.Path, nameprefix: str = ""):
        # parse time filename values
        pl_rois = pl.read_csv(filename)
        rois = []
        self.rois = {}
        for r in pl_rois.rows(named=True):
            if isinstance(r, dict):
                rois.append(r)
            elif isinstance(r, tuple):
                rois.append(r._asdict())
        for r in rois:
            participant = r["Participant"]
            self.rois[participant] = {}
            for k, v in r:
                if k != "Participant":
                    self.rois[participant][k] = self.parseDuration(v)
        self.name_prefix = nameprefix

    def split(
        self, sourcedrivedata: pydre.core.DriveData
    ) -> list[pydre.core.DriveData]:
        """
        return list of pydre.core.DriveData objects
        the 'roi' field of the objects will be filled with the roi tag listed
        in the roi definition file column name
        """
        timecol = "SimTime"
        output_list = []

        if sourcedrivedata.PartID in self.rois.keys():
            for roi, duration in self.rois[sourcedrivedata.PartID]:
                start, end = duration
                new_data = sliceByTime(start, end, timecol, sourcedrivedata.data)
                new_ddata = pydre.core.DriveData(
                    new_data, sourcedrivedata.sourcefilename
                )
                new_ddata.copyMetaData(sourcedrivedata)
                new_ddata.roi = roi
                output_list.append(new_ddata)
        return output_list

    def parseDuration(self, duration: str) -> tuple[float, float]:
        # parse a string indicating duration into a tuple of (starttime, endtime) in seconds
        # the string will have the format as:
        # time1-time2 where time1 or time 2 are either hr:min:sec or min:sec
        # example:  1:15:10-1:20:30
        # example : 02:32-08:45
        pair_regex = r"([\d:])-([\d:])"
        time_regex = r"(?:(\d+):)?(\d+):(\d+)"
        pair_result = re.match(pair_regex, duration)
        first_time_str, second_time_str = pair_result.group(1, 2)
        first_time_result = re.match(time_regex, first_time_str)
        second_time_result = re.match(time_regex, second_time_str)
        first_time = first_time_result.group(2) * 60 + first_time_result.group(3)
        if first_time_result.group(1):
            first_time += first_time_result.group(1) * 60 * 60
        second_time = second_time_result.group(2) * 60 + second_time_result.group(3)
        if second_time_result.group(1):
            second_time += second_time_result.group(1) * 60 * 60
        return (first_time, second_time)


class SpaceROI(ROIProcessor):
    x_column_name = "XPos"
    y_column_name = "YPos"

    def __init__(self, filename: str | pathlib.Path, nameprefix: str = ""):
        # parse time filename values
        # roi_info is a data frame containing the cutoff points for the region in each row.
        # It's columns must be roi, X1, X2, Y1, Y2
        pl_rois = pl.read_csv(filename, has_header=True)
        expected_columns = ["roi", "X1", "X2", "Y1", "Y2"]
        if not set(expected_columns).issubset(set(pl_rois.columns)):
            logger.error(
                f"SpaceROI file {filename} does not contain expected columns {expected_columns}"
            )
            raise ValueError
        # convert polars table into dictionary with 'roi' as the key and a dict as the value
        self.roi_info = pl_rois.rows_by_key("roi", unique=True, named=True)
        self.name_prefix = nameprefix

    def split(
        self, sourcedrivedata: pydre.core.DriveData
    ) -> list[pydre.core.DriveData]:
        return_list = []

        for roi_name, roi_location in self.roi_info.items():
            xmin = min(roi_location.get("X1"), roi_location.get("X2"))
            xmax = max(roi_location.get("X1"), roi_location.get("X2"))
            ymin = min(roi_location.get("Y1"), roi_location.get("Y2"))
            ymax = max(roi_location.get("Y1"), roi_location.get("Y2"))

            region_data = sourcedrivedata.data.filter(
                pl.col(self.x_column_name).cast(pl.Float32).is_between(xmin, xmax)
                & pl.col(self.y_column_name).cast(pl.Float32).is_between(ymin, ymax)
            )

            if region_data.height == 0:
                # try out PartID to get this to run cgw 5/20/2022
                # logger.warning("No data for SubjectID: {}, Source: {},  ROI: {}".format(
                #    ddata.SubjectID,
                #    ddata.sourcefilename,
                #    self.roi_info.roi[i]))
                logger.warning(
                    "No data for SubjectID: {}, Source: {},  ROI: {}".format(
                        ddata.PartID, ddata.sourcefilename, roi_name
                    )
                )
            else:
                # try out PartID to get this to run cgw 5/20/2022
                # logger.info("{} Line(s) read into ROI {} for Subject {} From file {}".format(
                #     len(region_data),
                #     self.roi_info.roi[i],
                #     ddata.SubjectID,
                #     ddata.sourcefilename))
                logger.info(
                    "{} Line(s) read into ROI {} for Subject {} From file {}".format(
                        region_data.height,
                        roi_name,
                        sourcedrivedata.PartID,
                        sourcedrivedata.sourcefilename,
                    )
                )
            new_ddata = pydre.core.DriveData(
                region_data, sourcedrivedata.sourcefilename
            )
            new_ddata.copyMetaData(sourcedrivedata)
            new_ddata.roi = roi_name
            return_list.append(new_ddata)

        return return_list


class ColumnROI(ROIProcessor):
    def __init__(self, columnname: str, nameprefix=""):
        # parse time filename values
        self.roi_column = columnname
        self.name_prefix = nameprefix

    def split(
        self, sourcedrivedata: pydre.core.DriveData
    ) -> Iterable[pydre.core.DriveData]:
        return_list = []

        for gname, gdata in sourcedrivedata.data.group_by(self.roi_column):
            gname = gname[0]
            if gname != pl.Null:
                new_ddata = pydre.core.DriveData(gdata, sourcedrivedata.sourcefilename)
                new_ddata.copyMetaData(sourcedrivedata)
                new_ddata.roi = str(gname)
                return_list.append(new_ddata)
        return return_list



import re

import polars as pl
from loguru import logger

import pydre.core
from pydre.filters import registerFilter


@registerFilter()
def modifyCriticalEventsCol(drivedata: pydre.core.DriveData):
    ident = drivedata.PartID
    ident_groups = re.match(r"(\d)(\d)(\d)(\d\d\d\d)[wW](\d)", ident)
    if ident_groups is None:
        logger.warning("Could not parse R2D ID " + ident)
        return [None]
    week = ident_groups.group(5)
    scenario = drivedata.scenarioName
    if week == "1" and scenario == "Load, Event":
        # between the x positions, change the critical event status to 1
        drivedata.data = drivedata.data.with_columns(
            pl.when(2165 < pl.col("XPos"), pl.col("XPos") < 2240)
            .then(1)
            .when(pl.col("CriticalEventStatus") == 1)
            .then(1)
            .otherwise(0)
            .alias("CriticalEventStatus")
        )
    return drivedata


@registerFilter()
def modifyUABdata(drivedata: pydre.core, headwaycutoff=50):
    ident = drivedata.PartID
    ident_groups = re.match(r"(\d)(\d)(\d)(\d\d\d\d)[wW](\d)", ident)
    if ident_groups is None:
        logger.warning("Could not parse R2D ID " + ident)
        return [None]
    location = ident_groups.group(2)
    df = drivedata.data
    if location == "1":
        # copy values from datTime into simTime
        df = df.with_columns(pl.col("DatTime").alias("SimTime"))
        # for files like Experimenter_3110007w1_No Load, Event_1665239271T-10-07-52.dat where the drive starts at
        # the end of a previous drive, trim the data leading up to the actual start
        df = df.with_columns(
            pl.col("XPos").cast(pl.Float32).diff().abs().alias("PosDiff")
        )
        df_actual_start = df.filter(df.get_column("PosDiff") > 500)
        if not df_actual_start.is_empty():
            start_time = df_actual_start.get_column("SimTime").item(0)
            df = df.filter(df.get_column("SimTime") > start_time)
        # modify xpos to match the starting value of dsl data
        start_pos = df.get_column("XPos").item(0)
        # add critical event status based on scenario type
        scenario = drivedata.scenarioName
        if scenario == "Load, Event":
            cutoff_df = df.filter(df.get_column("XPos") > (4350 + start_pos))
            try:
                start_cutoff = (
                    cutoff_df.filter(
                        cutoff_df.get_column("HeadwayDistance") < headwaycutoff
                    )
                    .get_column("XPos")
                    .item(0)
                ) + 135
                df = df.with_columns(
                    pl.when(
                        pl.col("XPos") > start_pos + 2155,
                        pl.col("XPos") < start_pos + 2239.5,
                    )
                    .then(1)
                    .when(
                        pl.col("XPos") > start_cutoff, pl.col("XPos") < start_pos + 4720
                    )
                    .then(1)
                    .when(
                        pl.col("XPos") > start_pos + 6191.4,
                        pl.col("XPos") < start_pos + 6242,
                    )
                    .then(1)
                    .otherwise(0)
                    .alias("CriticalEventStatus")
                )
            except IndexError:
                df = df.with_columns(
                    pl.when(
                        pl.col("XPos") > start_pos + 2155,
                        pl.col("XPos") < start_pos + 2239.5,
                    )
                    .then(1)
                    .when(
                        pl.col("XPos") > start_pos + 6191.4,
                        pl.col("XPos") < start_pos + 6242,
                    )
                    .then(1)
                    .otherwise(0)
                    .alias("CriticalEventStatus")
                )
        else:
            cutoff_df = df.filter(df.get_column("XPos") > (4550 + start_pos))
            try:
                start_cutoff = (
                    cutoff_df.filter(
                        cutoff_df.get_column("HeadwayDistance") < headwaycutoff
                    )
                    .get_column("XPos")
                    .item(0)
                ) + 135
                df = df.with_columns(
                    pl.when(
                        pl.col("XPos") > (start_pos + 1726),
                        pl.col("XPos") < (start_pos + 1790),
                    )
                    .then(1)
                    .when(
                        pl.col("XPos") > (start_pos + 3222),
                        pl.col("XPos") < (start_pos + 3300),
                    )
                    .then(1)
                    .when(
                        pl.col("XPos") > start_cutoff,
                        pl.col("XPos") < (start_pos + 5500),
                    )
                    .then(1)
                    .otherwise(0)
                    .alias("CriticalEventStatus")
                )
            except IndexError:
                df.with_columns(
                    pl.when(
                        pl.col("XPos") > (start_pos + 1726),
                        pl.col("XPos") < (start_pos + 1790),
                    )
                    .then(1)
                    .when(
                        pl.col("XPos") > (start_pos + 3222),
                        pl.col("XPos") < (start_pos + 3300),
                    )
                    .then(1)
                    .otherwise(0)
                    .alias("CriticalEventStatus")
                )
    drivedata.data = df
    return drivedata



from pathlib import Path
import pytest
import pydre.project
from loguru import logger

FIXTURE_DIR = Path(__file__).parent.resolve() / "test_data"


@pytest.mark.datafiles(FIXTURE_DIR / "good_projectfiles" / "test1_pf.json")
def test_project_loadjson(datafiles):
    proj = pydre.project.Project(datafiles / "test1_pf.json")
    assert isinstance(proj, pydre.project.Project)


@pytest.mark.datafiles(FIXTURE_DIR / "good_projectfiles" / "test1_pf.toml")
def test_project_loadtoml(datafiles):
    proj = pydre.project.Project(datafiles / "test1_pf.toml")
    assert isinstance(proj, pydre.project.Project)


def test_project_loadbadtoml():
    with pytest.raises(FileNotFoundError):
        proj = pydre.project.Project("doesnotexist.toml")


@pytest.mark.datafiles(
    FIXTURE_DIR / "good_projectfiles" / "test1_pf.json",
    FIXTURE_DIR / "good_projectfiles" / "test1_pf.toml",
)
def test_project_projequiv(datafiles):
    proj_json = pydre.project.Project(datafiles / "test1_pf.json")
    proj_toml = pydre.project.Project(datafiles / "test1_pf.toml")
    assert proj_json == proj_toml



import polars as pl
import pydre.core
from pydre.metrics import registerMetric


@registerMetric()
def averageBoxReactionTime(drivedata: pydre.core.DriveData):
    required_col = ["ReactionTime"]
    diff = drivedata.checkColumns(required_col)
    # Filter all reaction times that are negative (missed boxes) then output mean
    return (
        drivedata.data.filter(pl.col("ReactionTime") > 0)
        .select("ReactionTime")
        .mean()
        .item()
    )


@registerMetric()
def sdBoxReactionTime(drivedata: pydre.core.DriveData):
    required_col = ["ReactionTime"]
    diff = drivedata.checkColumns(required_col)
    return (
        drivedata.data.filter(pl.col("ReactionTime") > 0)
        .select("ReactionTime")
        .std()
        .item()
    )


def countBoxHits(drivedata: pydre.core.DriveData, cutoff=5):
    required_col = ["ReactionTime"]
    diff = drivedata.checkColumns(required_col)
    return drivedata.data.filter(
        pl.col("ReactionTime").is_between(0, cutoff, closed=None)
    ).height


def percentBoxHits(drivedata: pydre.core.DriveData, cutoff=5):
    required_col = ["ReactionTime"]
    diff = drivedata.checkColumns(required_col)
    df = drivedata.data.select(pl.col("ReactionTime"))
    return (
        df.filter(pl.col("ReactionTime").is_between(0, cutoff, closed=None)).height
        / df.height
    ) * 100


# negative reaction time from the filter indicates a missed box
def countBoxMisses(drivedata: pydre.core.DriveData):
    required_col = ["ReactionTime"]
    diff = drivedata.checkColumns(required_col)
    return drivedata.data.filter(pl.col("ReactionTime") < 0).height


def percentBoxMisses(drivedata: pydre.core.DriveData):
    required_col = ["ReactionTime"]
    diff = drivedata.checkColumns(required_col)
    df = drivedata.data.select(pl.col("ReactionTime"))
    return (df.filter(pl.col("ReactionTime") < 0).height / df.height) * 100


#
# @registerMetric()
# def boxMetrics(drivedata: pydre.core.DriveData, cutoff: float = 0, stat: str = "count"):
#     required_col = ["SimTime", "FeedbackButton", "BoxAppears"]
#     diff = drivedata.checkColumns(required_col)
#
#     total_boxclicks = pandas.Series(dtype="float64")
#     # original code here: total_boxclicks = pandas.Series()
#     # Got this warning on pandas 1.2.4: " DeprecationWarning: The default dtype for empty Series will be 'object'
#     # instead of 'float64' in a future version. Specify a dtype explicitly to silence this warning."
#     # Plz change it back to the original code if the current one leads to an issue
#     time_boxappeared = 0.0
#     time_buttonclicked = 0.0
#     hitButton = 0
#     df = pandas.DataFrame(drivedata.data, columns=required_col)  # drop other columns
#     df = pandas.DataFrame.drop_duplicates(
#         df.dropna(axis=0, how="any")
#     )  # remove nans and drop duplicates
#     if len(df) == 0:
#         return None
#     boxAppearsdf = df["BoxAppears"]
#     simTimedf = df["SimTime"]
#     boxOndf = boxAppearsdf.diff(1)
#     indicesBoxOn = boxOndf[boxOndf.values > 0.5].index[0:]
#     indicesBoxOff = boxOndf[boxOndf.values < 0.0].index[0:]
#     feedbackButtondf = df["FeedbackButton"]
#     reactionTimeList = list()
#     for counter in range(0, len(indicesBoxOn)):
#         boxOn = int(indicesBoxOn[counter])
#         boxOff = int(indicesBoxOff[counter])
#         startTime = simTimedf.loc[boxOn]
#         buttonClickeddf = feedbackButtondf.loc[boxOn:boxOff].diff(1)
#         buttonClickedIndices = buttonClickeddf[buttonClickeddf.values > 0.5].index[0:]
#
#         if len(buttonClickedIndices) > 0:
#             indexClicked = int(buttonClickedIndices[0])
#             clickTime = simTimedf.loc[indexClicked]
#             reactionTime = clickTime - startTime
#             reactionTimeList.append(reactionTime)
#         else:
#             if counter < len(indicesBoxOn) - 1:
#                 endIndex = counter + 1
#                 endOfBox = int(indicesBoxOn[endIndex])
#                 buttonClickeddf = feedbackButtondf.loc[boxOn : endOfBox - 1].diff(1)
#                 buttonClickedIndices = buttonClickeddf[
#                     buttonClickeddf.values > 0.5
#                 ].index[0:]
#                 if len(buttonClickedIndices) > 0:
#                     indexClicked = int(buttonClickedIndices[0])
#                     clickTime = simTimedf.loc[indexClicked]
#                     reactionTime = clickTime - startTime
#                     reactionTimeList.append(reactionTime)
#         sum = feedbackButtondf.loc[boxOn:boxOff].sum()
#         if sum > 0.000:
#             hitButton = hitButton + 1
#     if stat == "count":
#         return hitButton
#     elif stat == "mean":
#         mean = np.mean(reactionTimeList, axis=0)
#         return mean
#     elif stat == "sd":
#         sd = np.std(reactionTimeList, axis=0)
#         return sd
#     else:
#         print("Can't calculate that statistic.")
#     return hitButton



import pydre.core
import polars as pl
import pytest
import pydre.metrics
import pydre.metrics.common


def test_colMean():
    # Create a sample Polars DataFrame

    data = {"A": [1, 2, 3, 4, 5], "B": [10, 20, 30, 40, 50]}
    df = pl.DataFrame(data)
    dd = pydre.core.DriveData(df, "test.dat")

    # Test cases
    assert pydre.metrics.common.colMean(dd, var="A") == 3.0
    assert pydre.metrics.common.colMean(dd, var="B") == 30.0

    # Test cases with cutoff
    assert pydre.metrics.common.colMean(dd, var="A", cutoff=2.5) == 4.0
    assert pydre.metrics.common.colMean(dd, var="B", cutoff=25) == 40.0

    # Test for an invalid column name
    assert pydre.metrics.common.colMean(dd, var="InvalidColumn") is None


def test_colMedian():
    # Create a sample Polars DataFrame
    data = {"A": [-1, 2, 3, 4, 5], "B": [-10, 20, 30, 40, 50]}
    df = pl.DataFrame(data)
    dd = pydre.core.DriveData(df, "test.dat")

    # Test cases
    assert pydre.metrics.common.colMedian(dd, var="A") == 3.0
    assert pydre.metrics.common.colMedian(dd, var="B") == 30.0

    assert pydre.metrics.common.colMedian(dd, var="A", cutoff=2.5) == 4.0
    assert pydre.metrics.common.colMedian(dd, var="B", cutoff=25) == 40.0

    # Test for an invalid column name
    assert pydre.metrics.common.colMedian(dd, var="InvalidColumn") is None


def test_colSD():
    # Create a sample Polars DataFrame
    data = {"A": [1, 2, 3, 4, 5], "B": [10, 20, 30, 40, 50]}
    df = pl.DataFrame(data)
    dd = pydre.core.DriveData(df, "test.dat")

    # Test cases
    assert pytest.approx(pydre.metrics.common.colSD(dd, var="A"), 0.01) == 1.58
    assert pytest.approx(pydre.metrics.common.colSD(dd, var="B"), 0.01) == 15.8

    # Test for an invalid column name
    assert pydre.metrics.common.colSD(dd, var="InvalidColumn") is None


def test_colMax():
    # Create a sample Polars DataFrame
    data = {"A": [1, 2, 3, 4, 5], "B": [10, 20, 30, 40, 50]}
    df = pl.DataFrame(data)
    dd = pydre.core.DriveData(df, "test.dat")

    # Test cases
    assert pydre.metrics.common.colMax(dd, var="A") == 5
    assert pydre.metrics.common.colMax(dd, var="B") == 50

    # Test for an invalid column name
    assert pydre.metrics.common.colMax(dd, var="InvalidColumn") is None


def test_colMin():
    # Create a sample Polars DataFrame
    data = {"A": [1, 2, 3, 4, 5], "B": [10, 20, 30, 40, 50]}
    df = pl.DataFrame(data)
    dd = pydre.core.DriveData(df, "test.dat")

    # Test cases
    assert pydre.metrics.common.colMin(dd, var="A") == 1
    assert pydre.metrics.common.colMin(dd, var="B") == 10

    # Test for an invalid column name
    assert pydre.metrics.common.colMin(dd, var="InvalidColumn") is None


def test_timeAboveSpeed():
    # Create sample Polars DataFrames for testing
    data1 = {"SimTime": [0, 1, 2, 3, 4], "Velocity": [10, 20, 30, 40, 50]}
    df1 = pl.DataFrame(data1)
    dd1 = pydre.core.DriveData(df1, "test1.dat")

    data2 = {"SimTime": [0, 1, 2, 3], "Velocity": [5, 15, 25, 35]}  # Shorter duration
    df2 = pl.DataFrame(data2)
    dd2 = pydre.core.DriveData(df2, "test2.dat")

    # Test cases with various cutoffs and percentage options
    assert pydre.metrics.common.timeAboveSpeed(dd1) == 4
    assert (
        pydre.metrics.common.timeAboveSpeed(dd1, cutoff=30) == 3
    )  # 3 seconds above 30
    assert (
        pydre.metrics.common.timeAboveSpeed(dd1, cutoff=30.1) == 2
    )  # 2 seconds above 30.1
    assert pydre.metrics.common.timeAboveSpeed(dd1, cutoff=25, percentage=True) == (
        3 / 4
    )
    assert pydre.metrics.common.timeAboveSpeed(dd2, cutoff=20) == 2
    assert pydre.metrics.common.timeAboveSpeed(dd2, cutoff=40) == 0  # No time above 40

    # Test with a cutoff that results in no time above
    assert pydre.metrics.common.timeAboveSpeed(dd1, cutoff=60) == 0

    # Test with missing required columns
    data3 = {"SimTime": [0, 1, 2], "OtherColumn": [1, 2, 3]}
    df3 = pl.DataFrame(data3)
    dd3 = pydre.core.DriveData(df3, "test3.dat")
    assert pydre.metrics.common.timeAboveSpeed(dd3) is None

    # Test with non-numeric columns (assuming it raises ColumnsMatchError)
    data4 = {"SimTime": [0, 1, 2], "Velocity": ["slow", "medium", "fast"]}
    df4 = pl.DataFrame(data4)
    dd4 = pydre.core.DriveData(df4, "test4.dat")
    assert pydre.metrics.common.timeAboveSpeed(dd4) is None


def test_timeWithinSpeedLimit():
    # Create sample Polars DataFrames for testing
    data1 = {
        "SimTime": [0, 1, 2, 3, 4],
        "Velocity": [10, 12, 18, 20, 25],  # in meters per second
        "SpeedLimit": [35, 35, 35, 50, 50],  # in miles per hour
    }
    df1 = pl.DataFrame(data1)
    dd1 = pydre.core.DriveData(df1, "test1.dat")

    data2 = {"SimTime": [0, 1, 2], "Velocity": [5, 5, 25], "SpeedLimit": [30, 30, 60]}
    df2 = pl.DataFrame(data2)
    dd2 = pydre.core.DriveData(df2, "test2.dat")

    # Test cases with various lower limits and percentage options
    assert pydre.metrics.common.timeWithinSpeedLimit(dd1, lowerlimit=0) == 2
    assert (
        pydre.metrics.common.timeWithinSpeedLimit(dd1, lowerlimit=20, percentage=True)
        == 0.5
    )
    assert pydre.metrics.common.timeWithinSpeedLimit(dd2, lowerlimit=10) == 2
    assert pydre.metrics.common.timeWithinSpeedLimit(dd2, lowerlimit=20) == 1

    # Test with a lower limit that results in no time within limit
    assert pydre.metrics.common.timeWithinSpeedLimit(dd1, lowerlimit=100) == 0

    # Test with missing required columns
    data3 = {"SimTime": [0, 1, 2], "Velocity": [10, 20, 30]}
    df3 = pl.DataFrame(data3)
    dd3 = pydre.core.DriveData(df3, "test3.dat")
    assert pydre.metrics.common.timeWithinSpeedLimit(dd3) is None

    # Test with non-numeric columns
    data4 = {
        "SimTime": [0, 1, 2],
        "Velocity": ["slow", "medium", "fast"],
        "SpeedLimit": [30, 30, 40],
    }
    df4 = pl.DataFrame(data4)
    dd4 = pydre.core.DriveData(df4, "test4.dat")
    assert pydre.metrics.common.timeWithinSpeedLimit(dd4) is None



from typing import List

import numpy as np
import polars as pl
from loguru import logger

import pydre.core
from pydre.filters import registerFilter
from pydre.filters.common import mergeSplitFiletime, filetimeToDatetime

# The following functions need to be revised to work with polars rather than pandas

#
# @registerFilter()
# def smarteyeTimeSync(drivedata: pydre.core.DriveData, smarteye_vars: List[str]):
#     # TODO: switch to using polars
#
#     dt = drivedata.data.to_pandas()
#     # REALTIME_CLOCK is the 64-bit integer timestamp from SmartEye
#     # The clock data from SimObserver is in two different 32-bit integer values:
#     # hiFileTime encodes the high-order bits of the clock data
#     # lowFileTime encodes the low-order bits of the clock data
#     dt["SimCreatorClock"] = np.vectorize(mergeSplitFiletime)(
#         dt["hiFileTime"], dt["lowFileTime"]
#     )
#     dt["SimCreatorClock"] = dt["SimCreatorClock"].apply(filetimeToDatetime)
#     dt["SmartEyeClock"] = np.vectorize(filetimeToDatetime)(dt["REALTIME_CLOCK"])
#
#     # make a new data table with only the smarteye vars, for preparation for realignment
#     # first we should check if all the varables we want to shift are actually in the data
#     orig_columns = set(dt.columns)
#     smarteye_columns = set(smarteye_vars)
#
#     if not smarteye_columns.issubset(orig_columns):
#         logger.error(
#             "Some columns defined in the filter parameters are not in the DriveData: {}"
#             % [smarteye_columns - orig_columns]
#         )
#         # there is probably a cleaner way to do this operation.
#         # We want to keep only the smarteye data columns that are actually in the data file
#         smarteye_columns = smarteye_columns - (smarteye_columns - orig_columns)
#
#     smarteye_columns.add("SmartEyeClock")
#     smarteye_data = dt[smarteye_columns]
#     simcreator_data = dt[orig_columns - smarteye_columns]
#     drivedata.data = pl.from_pandas(
#         pandas.merge_asof(
#             simcreator_data,
#             smarteye_data,
#             left_on="SimCreatorClock",
#             right_on="SmartEyeClock",
#         )
#     )
#     return drivedata
#
#
# @registerFilter()
# def smoothGazeData(
#     drivedata: pydre.core.DriveData,
#     timeColName: str = "DatTime",
#     gazeColName: str = "FILTERED_GAZE_OBJ_NAME",
#     latencyShift=6,
# ):
#     # Right now this is just converting to a pandas DataFrame, doing the old computation and converting back
#     # TODO: convert to actually use polars
#
#     required_col = [timeColName, gazeColName]
#     diff = drivedata.checkColumns(required_col)
#
#     dt = drivedata.data.to_pandas()
#     cat_type = CategoricalDtype(
#         categories=[
#             "None",
#             "localCS.dashPlane",
#             "localCS.WindScreen",
#             "localCS.CSLowScreen",
#             "onroad",
#             "offroad",
#         ]
#     )
#     dt["gaze"] = dt[gazeColName].astype(cat_type)
#
#     # dt['gaze'].replace(['None', 'car.dashPlane', 'car.WindScreen'], ['offroad', 'offroad', 'onroad'], inplace=True)
#     dt["gaze"].replace(
#         ["None", "localCS.dashPlane", "localCS.WindScreen", "localCS.CSLowScreen"],
#         ["offroad", "offroad", "onroad", "offroad"],
#         inplace=True,
#     )
#
#     if len(dt["gaze"].unique()) < 2:
#         print("Bad gaze data, not enough variety. Aborting")
#         return None
#
#     # smooth frame blips
#     gaze_same = (
#         (dt["gaze"].shift(-1) == dt["gaze"].shift(1))
#         & (dt["gaze"].shift(-2) == dt["gaze"].shift(2))
#         & (dt["gaze"].shift(-2) == dt["gaze"].shift(-1))
#         & (dt["gaze"] != dt["gaze"].shift(1))
#     )
#     # print("{} frame blips".format(gaze_same.sum()))
#     dt.loc[gaze_same, "gaze"] = dt["gaze"].shift(-1)
#
#     # adjust for 100ms latency
#     dt["gaze"] = dt["gaze"].shift(-latencyShift)
#     dt["timedelta"] = pandas.to_timedelta(dt[timeColName].astype(float), unit="s")
#     dt.set_index("timedelta", inplace=True)
#
#     # filter out noise from the gaze column
#     # SAE J2396 defines fixations as at least 0.2 seconds,
#     min_delta = pandas.to_timedelta(0.2, unit="s")
#     # so we ignore changes in gaze that are less than that
#
#     # find list of runs
#     dt["gazenum"] = (dt["gaze"].shift(1) != dt["gaze"]).astype(int).cumsum()
#     n = dt["gazenum"].max()
#     dt = dt.reset_index()
#     # breakpoint()
#     durations = (
#         dt.groupby("gazenum")["timedelta"].max()
#         - dt.groupby("gazenum")["timedelta"].min()
#     )
#
#     # print("{} gazes before removing transitions".format(n))
#     short_gaze_count = 0
#     dt.set_index("gazenum")
#
#     short_duration_indices = durations[durations.lt(min_delta)].index.values
#     short_gaze_count = len(short_duration_indices)
#     dt.loc[dt["gazenum"].isin(short_duration_indices), "gaze"] = np.nan
#
#     # for x in trange(durations.index.min(), durations.index.max()):
#     # if durations.loc[x] < min_delta:
#     # short_gaze_count += 1
#     # dt['gaze'].where(dt['gazenum'] != x, inplace=True)
#     # dt.loc[x,'gaze']  = np.nan
#     # dt.loc[dt['gazenum'] == x, 'gaze'] = np.nan
#     dt.reset_index()
#     # print("{} transition gazes out of {} gazes total.".format(short_gaze_count, n))
#     dt["gaze"].fillna(method="bfill", inplace=True)
#     dt["gazenum"] = (dt["gaze"].shift(1) != dt["gaze"]).astype(int).cumsum()
#     # print("{} gazes after removing transitions.".format(dt['gazenum'].max()))
#
#     drivedata.data = polars.from_pandas(dt)
#     return drivedata



__all__ = ["common", "box", "driverdistraction"]

from functools import wraps
from typing import Optional, Callable
from loguru import logger
import pydre.core

metricsList: dict[str, Callable[[pydre.core.DriveData, ...], ...]] = {}
metricsColNames: dict[str, list[str]] = {}


def registerMetric(
    metricname: Optional[str] = None, columnnames: Optional[list[str]] = None
) -> Callable:
    def registering_decorator(
        func: Callable[[pydre.core.DriveData, ...], ...],
    ) -> Callable[[pydre.core.DriveData, ...], ...]:
        name = metricname
        if not metricname:
            name = func.__name__
        # register function
        metricsList[name] = func
        if columnnames:
            metricsColNames[name] = columnnames
        else:
            metricsColNames[name] = [
                name,
            ]
        return func

    return registering_decorator


def check_data_columns(arg):
    def argwrapper(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            logger.debug(f"{f} was called with arguments={args} and kwargs={kwargs}")
            value = f(*args, **kwargs)
            logger.debug(f"{f} return value {value}")
            return value

        return wrapper

    return argwrapper
